from django.contrib import admin

from .models import Rate


class RateAdmin(admin.ModelAdmin):
    list_display = ("currency", "value", "last_update", "backend")
    list_filter = ("currency",)
    ordering = ("currency",)
    readonly_fields = ("backend",)

    def last_update(self, instance):
        return instance.backend.last_update


admin.site.register(Rate, RateAdmin)



from django.core.exceptions import FieldDoesNotExist
from django.core.validators import MaxValueValidator, MinValueValidator
from django.utils.translation import gettext_lazy as _

from rest_framework.fields import empty
from rest_framework.serializers import DecimalField, ModelSerializer

from djmoney.models.fields import MoneyField as ModelField
from djmoney.models.validators import MaxMoneyValidator, MinMoneyValidator
from djmoney.money import Money
from djmoney.utils import MONEY_CLASSES, get_currency_field_name
from moneyed.classes import CurrencyDoesNotExist


class _PrimitiveMoney:
    """
    A container for ``Money`` data that does not do any validation of said data.
    It conveniently holds the amount and currency attributes to ease transformation into a valid ``Money`` instance.
    """

    __slots__ = ("amount", "currency")

    def __init__(self, amount, currency):
        self.amount = amount
        self.currency = currency


class MoneyField(DecimalField):
    """
    Treats ``Money`` objects as decimal values in representation and
    does decimal's validation during transformation to native value.
    """

    default_error_messages = {
        "invalid_currency": _("{currency!r} is not a valid currency"),
    }

    def __init__(self, *args, **kwargs):
        self.default_currency = kwargs.pop("default_currency", None)
        super().__init__(*args, **kwargs)
        # Rest Framework converts `min_value` / `max_value` to validators, that are not aware about `Money` class
        # We need to adjust them
        for idx, validator in enumerate(self.validators):
            if isinstance(validator, MinValueValidator):
                self.validators[idx] = MinMoneyValidator(self.min_value)
            elif isinstance(validator, MaxValueValidator):
                self.validators[idx] = MaxMoneyValidator(self.max_value)

    def to_representation(self, obj):
        """
        When ``field_currency`` is not in ``self.validated_data`` then ``obj`` is an instance of ``Decimal``, otherwise
        it is ``Money``.
        """
        if isinstance(obj, MONEY_CLASSES):
            obj = obj.amount
        return super().to_representation(obj)

    def to_internal_value(self, data):
        if isinstance(data, MONEY_CLASSES + (_PrimitiveMoney,)):
            amount = super().to_internal_value(data.amount)
            try:
                return Money(amount, data.currency)
            except CurrencyDoesNotExist:
                self.fail("invalid_currency", currency=data.currency)

        return super().to_internal_value(data)

    def get_value(self, data):
        parent_meta = getattr(self.parent, "Meta", None)

        default_currency = self.default_currency

        if parent_meta and hasattr(parent_meta, "model"):
            model = self.parent.Meta.model
            try:
                field = model._meta.get_field(self.source)
                default_currency = field.default_currency
            except FieldDoesNotExist as e:
                if not hasattr(model, self.source):
                    raise ValueError(
                        f"{self.source} is neither a db field nor a property on the model {model.__name__}"
                    ) from e

        amount = super().get_value(data)
        currency = data.get(get_currency_field_name(self.field_name), self.default_currency or default_currency)
        if currency and amount is not None and not isinstance(amount, MONEY_CLASSES) and amount is not empty:
            return _PrimitiveMoney(amount=amount, currency=currency)
        return amount


def register_money_field():
    ModelSerializer.serializer_field_mapping[ModelField] = MoneyField



from decimal import Decimal
from textwrap import dedent
from unittest.mock import patch

from django.core.exceptions import ImproperlyConfigured

import pytest

from djmoney.contrib.exchange.exceptions import MissingRate
from djmoney.contrib.exchange.models import _get_rate, convert_money, get_rate
from djmoney.money import Currency, Money


pytestmark = pytest.mark.django_db


@pytest.mark.parametrize(
    "source, target, expected, queries",
    (
        ("USD", "USD", 1, 0),
        ("USD", "EUR", 2, 1),
        ("EUR", "USD", Decimal("0.5"), 1),
        (Currency("USD"), "USD", 1, 0),
        ("USD", Currency("USD"), 1, 0),
    ),
)
@pytest.mark.usefixtures("simple_rates")
def test_get_rate(source, target, expected, django_assert_num_queries, queries):
    with django_assert_num_queries(queries):
        assert get_rate(source, target) == expected


@pytest.mark.parametrize(
    "source, target, expected",
    (
        ("NOK", "SEK", Decimal("1.067732610555839085161462146")),
        ("SEK", "NOK", Decimal("0.9365640705489186883886537319")),
    ),
)
@pytest.mark.usefixtures("default_openexchange_rates")
def test_rates_via_base(source, target, expected, django_assert_num_queries):
    with django_assert_num_queries(1):
        assert get_rate(source, target) == expected


@pytest.mark.parametrize("source, target", (("NOK", "ZAR"), ("ZAR", "NOK"), ("USD", "ZAR"), ("ZAR", "USD")))
@pytest.mark.usefixtures("default_openexchange_rates")
def test_unknown_currency_with_partially_exiting_currencies(source, target):
    with pytest.raises(MissingRate, match=f"Rate {source} \\-\\> {target} does not exist"):
        get_rate(source, target)


@pytest.mark.parametrize("source, target", (("USD", "EUR"), ("SEK", "ZWL")))
def test_unknown_currency(source, target):
    with pytest.raises(MissingRate, match=f"Rate {source} \\-\\> {target} does not exist"):
        get_rate(source, target)


def test_string_representation(backend):
    assert str(backend) == backend.name


@pytest.mark.usefixtures("simple_rates")
def test_cache():
    with patch("djmoney.contrib.exchange.models._get_rate", wraps=_get_rate) as original:
        assert get_rate("USD", "USD") == 1
        assert original.call_count == 0
        assert get_rate("USD", "EUR") == 2
        assert original.call_count == 1
        assert get_rate("USD", "EUR") == 2
        assert original.call_count == 1


def test_bad_configuration(settings):
    settings.INSTALLED_APPS.remove("djmoney.contrib.exchange")
    with pytest.raises(ImproperlyConfigured):
        convert_money(Money(1, "USD"), "EUR")
    settings.INSTALLED_APPS.append("djmoney.contrib.exchange")


def test_without_installed_exchange(testdir):
    """
    If there is no 'djmoney.contrib.exchange' in INSTALLED_APPS importing `Money` should not cause a RuntimeError.
    Details: GH-385.
    """
    testdir.mkpydir("money_app")
    testdir.makepyfile(
        app_settings="""
    DATABASES = {
        'default': {
            'ENGINE': 'django.db.backends.sqlite3',
            'NAME': 'test.db',
        }
    }
    INSTALLED_APPS = ['djmoney']
    SECRET_KEY = 'foobar'
    """
    )
    result = testdir.runpython_c(
        dedent(
            """
    import os
    os.environ['DJANGO_SETTINGS_MODULE'] = 'app_settings'
    from django import setup

    setup()
    from djmoney.money import Money
    print(Money(1, 'USD'))
    """
        )
    )
    result.stdout.fnmatch_lines(["$1.00"])



"""
Created on May 7, 2011

@author: jake
"""

from decimal import Decimal

from django.conf import settings
from django.core.validators import MinValueValidator
from django.db import models

from djmoney.models.fields import MoneyField
from djmoney.models.managers import money_manager, understands_money
from djmoney.models.validators import MaxMoneyValidator, MinMoneyValidator
from djmoney.money import Money
from moneyed import Money as OldMoney


# Import reversion if configured
if "reversion" in settings.INSTALLED_APPS:
    from reversion.revisions import register
else:
    register = lambda _: None  # noqa


def get_default():
    return Money("0.00", settings.TEST_DEFAULT_CURRENCY)


def get_default_currency():
    return settings.TEST_DEFAULT_CURRENCY


def get_currency_choices():
    a_list = ["DKK", "USD"]
    if settings.TEST_DEFAULT_CURRENCY not in a_list:
        a_list.append(settings.TEST_DEFAULT_CURRENCY)
    a_list.sort()
    return [(code, code) for code in a_list]


def get_min_value_validator():
    return Money(50, "DKK")


def get_max_value_validator():
    return Money(2000, "DKK")


class ModelWithVanillaMoneyField(models.Model):
    money = MoneyField(max_digits=10, decimal_places=2)
    second_money = MoneyField(max_digits=10, decimal_places=2, default=0.0, default_currency="EUR")
    integer = models.IntegerField(default=0)


class ModelWithNullableCurrency(models.Model):
    money = MoneyField(max_digits=10, decimal_places=2, null=True, default_currency=None)


class ModelWithDefaultAsInt(models.Model):
    money = MoneyField(default=123, max_digits=10, decimal_places=2, default_currency="GHS")


class ModelWithUniqueIdAndCurrency(models.Model):
    money = MoneyField(default=123, max_digits=10, decimal_places=2, default_currency="GHS")

    class Meta:
        unique_together = ("id", "money")


class ModelWithDefaultAsStringWithCurrency(models.Model):
    money = MoneyField(default="123 USD", max_digits=10, decimal_places=2)

    class Meta:
        verbose_name = "model_default_string_currency"


class ModelWithDefaultAsString(models.Model):
    money = MoneyField(default="123", max_digits=10, decimal_places=2, default_currency="PLN")


class ModelWithDefaultAsFloat(models.Model):
    money = MoneyField(default=12.05, max_digits=10, decimal_places=2, default_currency="PLN")


class ModelWithDefaultAsDecimal(models.Model):
    money = MoneyField(default=Decimal("0.01"), max_digits=10, decimal_places=2, default_currency="CHF")


class ModelWithDefaultAsMoney(models.Model):
    money = MoneyField(default=Money("0.01", "RUB"), max_digits=10, decimal_places=2)


class ModelWithDefaultAsOldMoney(models.Model):
    money = MoneyField(default=OldMoney("0.01", "RUB"), max_digits=10, decimal_places=2)


class ModelWithTwoMoneyFields(models.Model):
    amount1 = MoneyField(max_digits=10, decimal_places=2)
    amount2 = MoneyField(max_digits=10, decimal_places=3)


class ModelRelatedToModelWithMoney(models.Model):
    moneyModel = models.ForeignKey(ModelWithVanillaMoneyField, on_delete=models.CASCADE)


class ModelWithChoicesMoneyField(models.Model):
    money = MoneyField(max_digits=10, decimal_places=2, currency_choices=[("USD", "US Dollars"), ("ZWN", "Zimbabwian")])


class ModelWithNonMoneyField(models.Model):
    money = MoneyField(max_digits=10, decimal_places=2, default=0.0, default_currency="USD")
    desc = models.CharField(max_length=10)


class AbstractModel(models.Model):
    money = MoneyField(max_digits=10, decimal_places=2, default=0.0, default_currency="USD")
    m2m_field = models.ManyToManyField(ModelWithDefaultAsInt)

    class Meta:
        abstract = True


class InheritorModel(AbstractModel):
    second_field = MoneyField(max_digits=10, decimal_places=2, default=0.0, default_currency="USD")


class RevisionedModel(models.Model):
    amount = MoneyField(max_digits=10, decimal_places=2, default=0.0, default_currency="USD")


register(RevisionedModel)


class BaseModel(models.Model):
    money = MoneyField(max_digits=10, decimal_places=2, default=0.0, default_currency="USD")


class InheritedModel(BaseModel):
    second_field = MoneyField(max_digits=10, decimal_places=2, default=0.0, default_currency="USD")


class SimpleModel(models.Model):
    money = MoneyField(max_digits=10, decimal_places=2, default=0.0, default_currency="USD")


class NullMoneyFieldModel(models.Model):
    field = MoneyField(max_digits=10, decimal_places=2, null=True, default_currency="USD", blank=True)


class ModelManager(models.Manager):
    pass


class MoneyFieldModelWithProperty(models.Model):
    money = MoneyField(max_digits=10, decimal_places=2, default=0.0, default_currency="USD")

    @property
    def ten_extra_monies(self):
        return self.money + Money(10, "USD")


class NotNullMoneyFieldModel(models.Model):
    money = MoneyField(max_digits=10, decimal_places=2)

    objects = money_manager(ModelManager())


class ProxyModelWrapper(NotNullMoneyFieldModel):
    class Meta:
        proxy = True


class ProxyModel(SimpleModel):
    class Meta:
        proxy = True


class MoneyManager(models.Manager):
    @understands_money
    def super_method(self, **kwargs):
        return self.filter(**kwargs)


class ModelWithCustomManager(models.Model):
    field = MoneyField(max_digits=10, decimal_places=2)

    manager = money_manager(MoneyManager())


class DateTimeModel(models.Model):
    field = MoneyField(max_digits=10, decimal_places=2)
    created = models.DateTimeField(null=True, blank=True)


class ModelIssue300(models.Model):
    money = models.ForeignKey(DateTimeModel, on_delete=models.CASCADE)
    price = MoneyField(max_digits=10, decimal_places=2, default_currency="EUR", default=Decimal("0.0"))


class ModelWithValidation(models.Model):
    balance = MoneyField(max_digits=10, decimal_places=2, validators=[MinValueValidator(Money(100, "GBP"))])


class ModelWithSharedCurrency(models.Model):
    first = MoneyField(max_digits=10, decimal_places=2, currency_field_name="currency")
    second = MoneyField(max_digits=10, decimal_places=2, currency_field_name="currency")


class ValidatedMoneyModel(models.Model):
    money = MoneyField(
        max_digits=10,
        decimal_places=2,
        validators=[
            MinMoneyValidator({"EUR": 100, "USD": 50}),
            MaxMoneyValidator({"EUR": 1000, "USD": 500}),
            MinMoneyValidator(Money(500, "NOK")),
            MaxMoneyValidator(Money(900, "NOK")),
            MinMoneyValidator(get_min_value_validator),
            MaxMoneyValidator(get_max_value_validator),
            MinMoneyValidator(10),
            MaxMoneyValidator(1500),
        ],
    )


class PositiveValidatedMoneyModel(models.Model):
    """Validated model with a field requiring a non-negative value."""

    money = MoneyField(max_digits=10, decimal_places=2, validators=[MinMoneyValidator(0)])


class ModelWithCustomDefaultManager(models.Model):
    field = MoneyField(max_digits=10, decimal_places=2)

    custom = models.Manager()

    class Meta:
        default_manager_name = "custom"


class CryptoModel(models.Model):
    money = MoneyField(max_digits=10, decimal_places=2, currency_max_length=4)


class PreciseModel(models.Model):
    money = MoneyField(max_digits=10, decimal_places=4)


class ModelWithDefaultPrecision(models.Model):
    money = MoneyField(max_digits=10)


class ModelWithNullDefaultOnNonNullableField(models.Model):
    money = MoneyField(max_digits=10, decimal_places=2, default=None, default_currency=None)


class ModelWithCallableDefault(models.Model):

    money = MoneyField(max_digits=10, decimal_places=2, default=get_default, blank=True)


class ModelWithCallableDefaultCurrency(models.Model):

    money = MoneyField(max_digits=10, decimal_places=2, default_currency=get_default_currency)


class ModelWithCallableDefaultAndDefaultCurrency(models.Model):

    money = MoneyField(
        max_digits=10, decimal_places=2, default=get_default, default_currency=get_default_currency, blank=True
    )


class ModelWithCallableCurrencyChoices(models.Model):

    money = MoneyField(
        max_digits=10, decimal_places=2, default_currency=get_default_currency, currency_choices=get_currency_choices
    )


class ParentModel(models.Model):
    pass


class ModelWithParentAndCallableFields(ModelWithCallableDefaultAndDefaultCurrency):
    parent = models.ForeignKey(ParentModel, on_delete=models.CASCADE)



import json
import sys

import django
from django.core.serializers.base import DeserializationError
from django.core.serializers.json import Serializer as JSONSerializer

from djmoney.money import Money

from .models.fields import MoneyField
from .utils import get_currency_field_name


Serializer = JSONSerializer


def Deserializer(stream_or_string, **options):  # noqa
    """
    Deserialize a stream or string of JSON data.
    """
    # Local imports to allow using modified versions of `_get_model` / `_get_model_from_node`
    # It could be patched in runtime via `unittest.mock.patch` for example
    from django.core.serializers.python import Deserializer as PythonDeserializer

    if django.VERSION < (5, 2):
        from django.core.serializers.python import _get_model
    else:
        _get_model = PythonDeserializer._get_model_from_node

    ignore = options.pop("ignorenonexistent", False)

    if not isinstance(stream_or_string, (bytes, str)):
        stream_or_string = stream_or_string.read()
    if isinstance(stream_or_string, bytes):
        stream_or_string = stream_or_string.decode("utf-8")
    try:
        for obj in json.loads(stream_or_string):
            try:
                Model = _get_model(obj["model"])
            except DeserializationError:
                if ignore:
                    continue
                else:
                    raise
            money_fields = {}
            fields = {}
            field_names = {field.name for field in Model._meta.get_fields()}
            for field_name, field_value in obj["fields"].items():
                if ignore and field_name not in field_names:
                    # skip fields no longer on model
                    continue
                field = Model._meta.get_field(field_name)
                if isinstance(field, MoneyField) and field_value is not None:
                    money_fields[field_name] = Money(
                        field_value, obj["fields"][get_currency_field_name(field_name, field)]
                    )
                else:
                    fields[field_name] = field_value
            obj["fields"] = fields

            for inner_obj in PythonDeserializer([obj], **options):
                for field, value in money_fields.items():
                    setattr(inner_obj.object, field, value)
                yield inner_obj
    except (GeneratorExit, DeserializationError):
        raise
    except Exception as exc:
        raise DeserializationError.with_traceback(DeserializationError(exc), sys.exc_info()[2])



import json
import ssl
from decimal import Decimal
from typing import Optional
from urllib.parse import parse_qsl, urlparse, urlunparse
from urllib.request import Request, urlopen

from django.db.transaction import atomic
from django.utils.http import urlencode

from djmoney import settings

from ..models import ExchangeBackend, Rate


try:
    import certifi
except ImportError:
    raise ImportError("Please install dependency certifi - pip install certifi")


class BaseExchangeBackend:
    name: Optional[str] = None
    url: Optional[str] = None

    def get_rates(self, **kwargs):
        """
        Returns a mapping <currency>: <rate>.
        """
        raise NotImplementedError

    def get_url(self, **params):
        """
        Updates base url with provided GET parameters.
        """
        parts = list(urlparse(self.url))
        query = dict(parse_qsl(parts[4]))
        query.update(params)
        parts[4] = urlencode(query)
        return urlunparse(parts)

    def get_params(self):
        """
        Default GET parameters for the request.
        """
        return {}

    def get_response(self, **params):
        headers = {
            "User-Agent": "Mozilla/5.0",
        }
        request = Request(self.get_url(**params), headers=headers)
        context = ssl.create_default_context(cafile=certifi.where())
        response = urlopen(request, context=context)
        return response.read()

    def parse_json(self, response):
        if isinstance(response, bytes):
            response = response.decode("utf-8")
        return json.loads(response, parse_float=Decimal)

    @atomic
    def update_rates(self, base_currency=settings.BASE_CURRENCY, **kwargs):
        """
        Updates rates for the given backend.
        """
        backend, _ = ExchangeBackend.objects.update_or_create(name=self.name, defaults={"base_currency": base_currency})
        params = self.get_params()
        params.update(base_currency=base_currency, **kwargs)
        Rate.objects.bulk_create(
            [
                Rate(currency=currency, value=value, backend=backend)
                for currency, value in self.get_rates(**params).items()
            ],
            update_conflicts=True,
            unique_fields=["currency", "backend"],
            update_fields=["value"],
        )


class SimpleExchangeBackend(BaseExchangeBackend):
    """
    Simple backend implementation.
    Assumes JSON response with `rates` key inside.
    """

    def get_rates(self, **params):
        response = self.get_response(**params)
        return self.parse_json(response)["rates"]



import sys

from django.core.exceptions import ImproperlyConfigured

import pytest

from djmoney.contrib.exchange.backends import FixerBackend, OpenExchangeRatesBackend, UniRateBackend
from djmoney.contrib.exchange.models import ExchangeBackend, Rate, get_rate

from .conftest import ExchangeTest, FixedOneBackend, FixedTwoBackend


pytestmark = pytest.mark.django_db


class TestBackends(ExchangeTest):
    def test_get_rates(self):
        assert self.backend.get_rates() == self.expected

    def test_initial_update_rates(self):
        self.backend.update_rates()
        self.assert_rates()

    def test_second_update_rates(self):
        self.backend.update_rates()
        backend = ExchangeBackend.objects.get(name=self.backend.name)
        last_update = backend.last_update
        self.backend.update_rates()
        backend.refresh_from_db()
        assert last_update < backend.last_update


@pytest.mark.parametrize("backend", (FixerBackend, OpenExchangeRatesBackend, UniRateBackend))
def test_missing_settings(backend):
    with pytest.raises(ImproperlyConfigured):
        backend(access_key=None)


@pytest.mark.usefixtures("two_backends_data")
def test_two_backends():
    """
    Two different backends should not interfere with each other.
    """
    for backend in (FixedOneBackend, FixedTwoBackend):
        assert Rate.objects.filter(backend__name=backend.name).count() == 1
    assert get_rate("USD", "EUR", backend=FixedOneBackend.name) == 1
    assert get_rate("USD", "EUR", backend=FixedTwoBackend.name) == 2


@pytest.fixture
def no_certifi():
    certifi = sys.modules["certifi"]
    sys.modules["certifi"] = None
    base = sys.modules.pop("djmoney.contrib.exchange.backends.base")
    yield
    sys.modules["certifi"] = certifi
    # Return the old module to keep all existing mocks in other tests
    sys.modules["djmoney.contrib.exchange.backends.base"] = base


@pytest.mark.usefixtures("no_certifi")
def test_certifi_not_installed():
    with pytest.raises(ImportError, match="Please install dependency certifi - pip install certifi"):
        __import__("djmoney.contrib.exchange.backends.base")



import pytest

from djmoney.contrib.exchange.models import ExchangeBackend, Rate, get_default_backend_name
from djmoney.money import Money
from tests.testapp.models import InheritorModel, ModelWithDefaultAsInt


@pytest.fixture
def m2m_object():
    return ModelWithDefaultAsInt.objects.create(money=Money(100, "USD"))


@pytest.fixture
def backend():
    return ExchangeBackend.objects.create(name=get_default_backend_name(), base_currency="USD")


@pytest.fixture
def autoconversion(backend, settings):
    settings.AUTO_CONVERT_MONEY = True
    Rate.objects.create(currency="EUR", value="0.88", backend=backend)


@pytest.fixture
def concrete_instance(m2m_object):
    instance = InheritorModel.objects.create()
    instance.m2m_field.add(m2m_object)
    return instance


pytest_plugins = "pytester"



from django.core.exceptions import ImproperlyConfigured

from djmoney import settings

from .base import SimpleExchangeBackend


class FixerBackend(SimpleExchangeBackend):
    name = "fixer.io"

    def __init__(self, url=settings.FIXER_URL, access_key=settings.FIXER_ACCESS_KEY):
        if access_key is None:
            raise ImproperlyConfigured("settings.FIXER_ACCESS_KEY should be set to use FixerBackend")
        self.url = url
        self.access_key = access_key

    def get_params(self):
        # support both `data.fixer.io` and `api.apilayer.com` auth params
        return {
            "apikey": self.access_key,
            "access_key": self.access_key,
        }



import json
from contextlib import contextmanager
from decimal import Decimal
from unittest.mock import Mock, patch

from django.core.cache import cache

import pytest

from djmoney.contrib.exchange.backends import FixerBackend, OpenExchangeRatesBackend, UniRateBackend
from djmoney.contrib.exchange.backends.base import BaseExchangeBackend
from djmoney.contrib.exchange.models import ExchangeBackend, Rate


pytestmarks = pytest.mark.django_db


OPEN_EXCHANGE_RATES_RESPONSE = """{
    "disclaimer": "Usage subject to terms: https://openexchangerates.org/terms",
    "license": "https://openexchangerates.org/license",
    "timestamp": 1522749600,
    "base": "USD",
    "rates": {"EUR": 0.812511, "NOK": 7.847505, "SEK": 8.379037}
}"""
OPEN_EXCHANGE_RATES_EXPECTED = json.loads(OPEN_EXCHANGE_RATES_RESPONSE, parse_float=Decimal)["rates"]

FIXER_RESPONSE = """{
    "success":true,
    "timestamp":1522788248,
    "base":"EUR",
    "date":"2018-04-03",
    "rates":{"USD":1.227439,"NOK":9.624334,"SEK":10.300293}
}"""
FIXER_EXPECTED = json.loads(FIXER_RESPONSE, parse_float=Decimal)["rates"]

UNIRATE_RESPONSE = """{
    "amount": 1.0,
    "base": "USD",
    "rates": {"EUR": 0.85, "NOK": 10.5, "SEK": 11.0}
}"""
UNIRATE_EXPECTED = json.loads(UNIRATE_RESPONSE, parse_float=Decimal)["rates"]


@contextmanager
def mock_backend(value):
    response = Mock()
    response.read.return_value = value
    with patch("djmoney.contrib.exchange.backends.base.urlopen", return_value=response):
        yield


class ExchangeTest:
    @pytest.fixture(
        autouse=True,
        params=(
            (OpenExchangeRatesBackend, OPEN_EXCHANGE_RATES_RESPONSE, OPEN_EXCHANGE_RATES_EXPECTED),
            (FixerBackend, FIXER_RESPONSE, FIXER_EXPECTED),
            (UniRateBackend, UNIRATE_RESPONSE, UNIRATE_EXPECTED),
        ),
    )
    def setup(self, request):
        klass, response_value, expected = request.param
        self.backend = klass()
        self.expected = expected
        with mock_backend(response_value):
            yield

    def assert_rates(self):
        backend = ExchangeBackend.objects.get()
        assert Rate.objects.count() == len(self.expected)
        for currency, rate in self.expected.items():
            assert Rate.objects.filter(currency=currency, value=rate, backend=backend)


class FixedOneBackend(BaseExchangeBackend):
    name = "first"

    def get_rates(self, **params):
        return {"EUR": 1}


class FixedTwoBackend(BaseExchangeBackend):
    name = "second"

    def get_rates(self, **params):
        return {"EUR": 2}


@pytest.fixture
def two_backends_data():
    FixedOneBackend().update_rates()
    FixedTwoBackend().update_rates()


@pytest.fixture
def simple_rates(backend):
    Rate.objects.create(currency="EUR", value=2, backend=backend)


@pytest.fixture
def default_openexchange_rates():
    with mock_backend(OPEN_EXCHANGE_RATES_RESPONSE):
        OpenExchangeRatesBackend().update_rates()


@pytest.fixture(autouse=True)
def django_cache():
    cache.clear()
    yield



import operator
from types import MappingProxyType
from typing import Optional

from django.conf import settings

from moneyed import CURRENCIES, XXX, Currency


# The default currency, you can define this in your project's settings module
# This has to be a currency object imported from moneyed
DEFAULT_CURRENCY: Optional[Currency] = getattr(settings, "DEFAULT_CURRENCY", None)


# The default currency choices, you can define this in your project's
# settings module
PROJECT_CURRENCIES = getattr(settings, "CURRENCIES", None)
CURRENCY_CHOICES = getattr(settings, "CURRENCY_CHOICES", None)

if CURRENCY_CHOICES is None:
    if PROJECT_CURRENCIES:
        CURRENCY_CHOICES = [(code, CURRENCIES[code].name) for code in PROJECT_CURRENCIES]
    else:
        # 'XXX' is a backport from py-moneyed for transactions involving no currency;
        # this was previously referred to as moneyed.classes.DEFAULT_CURRENCY_CODE
        CURRENCY_CHOICES = [(c.code, c.name) for i, c in CURRENCIES.items() if c != XXX]

CURRENCY_CHOICES.sort(key=operator.itemgetter(1, 0))
DECIMAL_PLACES = getattr(settings, "CURRENCY_DECIMAL_PLACES", 2)

OPEN_EXCHANGE_RATES_URL = getattr(settings, "OPEN_EXCHANGE_RATES_URL", "https://openexchangerates.org/api/latest.json")
OPEN_EXCHANGE_RATES_APP_ID = getattr(settings, "OPEN_EXCHANGE_RATES_APP_ID", None)
FIXER_URL = getattr(settings, "FIXER_URL", "http://data.fixer.io/api/latest")
FIXER_ACCESS_KEY = getattr(settings, "FIXER_ACCESS_KEY", None)
UNIRATE_URL = getattr(settings, "UNIRATE_URL", "https://api.unirateapi.com/api/rates")
UNIRATE_ACCESS_KEY = getattr(settings, "UNIRATE_ACCESS_KEY", None)
BASE_CURRENCY = getattr(settings, "BASE_CURRENCY", "USD")
EXCHANGE_BACKEND = getattr(settings, "EXCHANGE_BACKEND", "djmoney.contrib.exchange.backends.OpenExchangeRatesBackend")
RATES_CACHE_TIMEOUT = getattr(settings, "RATES_CACHE_TIMEOUT", 600)

CURRENCY_CODE_MAX_LENGTH = getattr(settings, "CURRENCY_CODE_MAX_LENGTH", 3)

MONEY_FORMAT = MappingProxyType(getattr(settings, "MONEY_FORMAT", {}))



from django.core.management import call_command

import pytest

from djmoney.contrib.exchange.models import Rate

from .conftest import ExchangeTest, FixedOneBackend, FixedTwoBackend


pytestmark = pytest.mark.django_db

BACKEND_PATH = FixedOneBackend.__module__ + "." + FixedOneBackend.__name__


class TestCommand(ExchangeTest):
    def test_update_rates(self):
        call_command("update_rates")
        self.assert_rates()


def test_custom_backend():
    call_command("update_rates", backend=BACKEND_PATH)
    assert Rate.objects.filter(currency="EUR", value=1).exists()


@pytest.mark.usefixtures("two_backends_data")
class TestClearRates:
    def test_for_specific_backend(self):
        call_command("clear_rates", backend=BACKEND_PATH)
        assert not Rate.objects.filter(backend=FixedOneBackend.name).exists()
        assert Rate.objects.filter(backend=FixedTwoBackend.name).exists()

    def test_for_all(self):
        call_command("clear_rates", all=True)
        assert not Rate.objects.exists()



from django.utils.translation import override

import pytest

from djmoney.money import DefaultMoney, Money, get_current_locale


def test_repr():
    assert repr(Money("10.5", "USD")) == "Money('10.5', 'USD')"


def test_legacy_repr():
    assert repr(Money("10.5", "USD")) == "Money('10.5', 'USD')"


def test_html_safe():
    assert Money("10.5", "EUR").__html__() == "€10.50"


def test_html_unsafe():
    class UnsafeMoney(Money):
        def __str__(self):
            return "<script>"

    assert UnsafeMoney(100, "USD").__html__() == "&lt;script&gt;"


def test_default_mul():
    assert Money(10, "USD") * 2 == Money(20, "USD")


def test_default_truediv():
    assert Money(10, "USD") / 2 == Money(5, "USD")
    assert Money(10, "USD") / Money(2, "USD") == 5


def test_reverse_truediv_fails():
    with pytest.raises(TypeError):
        10 / Money(5, "USD")


@pytest.mark.parametrize(
    "locale, expected",
    (
        ("pl", "pl"),
        ("pl-pl", "pl_PL"),
        ("sv", "sv"),
        ("sv-se", "sv_SE"),
        ("en-us", "en_US"),
        ("en-gb", "en_GB"),
    ),
)
def test_get_current_locale(locale, expected):
    with override(locale):
        assert get_current_locale() == expected


def test_round():
    assert round(Money("1.69", "USD"), 1) == Money("1.7", "USD")


def test_configurable_decimal_number():
    # Override default configuration per instance, keeps human readable output to default
    mny = Money("10.543", "USD", decimal_places=3)
    assert str(mny) == "$10.54"
    assert mny.decimal_places == 3


def test_localize_defaults_to_currency_decimal_places():
    assert str(Money("10.543125", "USD", decimal_places=5)) == "$10.54"


def test_add_decimal_places():
    one = Money("1.0000", "USD", decimal_places=4)
    two = Money("2.000000005", "USD", decimal_places=10)

    result = one + two
    assert result.decimal_places == 10


def test_add_decimal_places_zero():
    two = Money("2.005", "USD", decimal_places=3)

    result = two + 0
    assert result.decimal_places == 3


@pytest.mark.parametrize("decimal_places", (1, 4))
@pytest.mark.parametrize(
    "operation",
    (
        lambda a, d: a * 2,
        lambda a, d: 2 * a,
        lambda a, d: a / 5,
        lambda a, d: a - Money("2", "USD", decimal_places=d),
        lambda a, d: Money("2", "USD", decimal_places=d) - a,
        lambda a, d: a + Money("2", "USD", decimal_places=d),
        lambda a, d: Money("2", "USD", decimal_places=d) + a,
        lambda a, d: -a,
        lambda a, d: +a,
        lambda a, d: abs(a),
        lambda a, d: 5 % a,
        lambda a, d: round(a),
        lambda a, d: a.round(),
    ),
)
def test_keep_decimal_places(operation, decimal_places):
    # Arithmetic operations should keep the `decimal_places` value
    amount = Money("1.0000", "USD", decimal_places=decimal_places)
    new = operation(amount, decimal_places)
    assert new.decimal_places == decimal_places


def test_sub_negative():
    # See GH-593
    total = DefaultMoney(0, "EUR")
    bills = (Money(8, "EUR"), Money(25, "EUR"))
    for bill in bills:
        total -= bill
    assert total == Money(-33, "EUR")


@pytest.mark.parametrize("decimal_places", [None, 0, 1, 4])
def test_proper_copy_of_attributes(decimal_places):
    one = Money(1, "EUR")

    assert one.decimal_places == 2, "default value"

    two = Money(2, "EUR", decimal_places=decimal_places)

    assert two.decimal_places == decimal_places

    result = Money(3, "EUR")
    one._copy_attributes(two, result)

    assert result.decimal_places == max(2, decimal_places) if decimal_places is not None else 2

    result = Money(0, "EUR")
    one._copy_attributes(Money(1, "EUR"), result)

    result = Money(0, "EUR")
    one._copy_attributes(1, result)

    assert result.decimal_places == 2

    result = Money(0, "EUR")
    two._copy_attributes(1, result)

    assert result.decimal_places == decimal_places if decimal_places else 2



from django.forms import HiddenInput, MultiWidget, Select, TextInput

from ..settings import CURRENCY_CHOICES


__all__ = (
    "HiddenMoneyWidget",
    "MoneyWidget",
)


class HiddenMoneyWidget(HiddenInput):

    def format_value(self, value):
        if value is None:
            return ""
        # For hidden inputs, we are sometimes supplied with a pre-formatted value
        if isinstance(value, str):
            return value
        # The hidden input widget is used in a MultiField (MoneyField!)
        if isinstance(value, list):
            amount, currency = value
            return f"{amount} {currency}"
        # Assuming it's a money object
        return f"{value.amount} {value.currency}"


class MoneyWidget(MultiWidget):
    def __init__(
        self,
        choices=CURRENCY_CHOICES,
        amount_widget=TextInput,
        currency_widget=None,
        default_currency=None,
        *args,
        **kwargs,
    ):
        self.default_currency = default_currency
        if not currency_widget:
            currency_widget = Select(choices=choices)
        widgets = (amount_widget, currency_widget)
        super().__init__(widgets, *args, **kwargs)

    def decompress(self, value):
        if value is not None:
            # Value provided by hidden input submissions somehow being evaluated in this widget
            # TBD: Why does that happen when MoneyField.hidden_widget() is defined?
            if isinstance(value, str):
                if " " in value:
                    amount, currency = value.split(" ")
                    return [amount, currency]
                raise ValueError(f"Invalid money value: {value}")
            if isinstance(value, (list, tuple)):
                return value
            return [value.amount, value.currency]
        return [None, self.default_currency]



from functools import wraps

from django.core.exceptions import FieldDoesNotExist
from django.db.models import NOT_PROVIDED, Case, F, Q
from django.db.models.constants import LOOKUP_SEP
from django.db.models.expressions import BaseExpression
from django.db.models.functions import Cast, Coalesce
from django.utils.encoding import smart_str

from ..utils import MONEY_CLASSES, get_currency_field_name, prepare_expression
from .fields import CurrencyField, MoneyField


def _get_clean_name(model, name):
    # Get rid of __lt, __gt etc for the currency lookup
    if LOOKUP_SEP not in name:
        return name
    lookup_fields = name.split(LOOKUP_SEP)
    field = _get_field(model, name)
    return name.rsplit(LOOKUP_SEP, lookup_fields.index(field.name) + 1)[0]


def _get_field(model, name):
    lookup_fields = name.split(LOOKUP_SEP)
    prev_field = None
    opts = model._meta
    for field_name in lookup_fields:
        if field_name == "pk":
            field_name = opts.pk.name
        try:
            field = opts.get_field(field_name)
        except FieldDoesNotExist:
            # Ignore valid query lookups.
            if prev_field and prev_field.get_lookup(field_name):
                continue
        else:
            prev_field = field
            if hasattr(field, "get_path_info"):
                # This field is a relation, update opts to follow the relation
                path_info = field.get_path_info()
                opts = path_info[-1].to_opts
    return prev_field


def is_in_lookup(name, value):
    return hasattr(value, "__iter__") & (name.split(LOOKUP_SEP)[-1] == "in")


def _convert_in_lookup(model, field_name, options):
    """
    ``in`` lookup can not be represented as keyword lookup.
    It requires transformation to combination of ``Q`` objects.

    Example:

        amount__in=[Money(10, 'EUR'), Money(5, 'USD)]

        is equivalent to:

        Q(amount=10, amount_currency='EUR') or Q(amount=5, amount_currency='USD')
    """
    field = _get_field(model, field_name)
    new_query = Q()
    for value in options:
        if isinstance(value, MONEY_CLASSES):
            # amount__in=[Money(1, 'EUR'), Money(2, 'EUR')]
            option = {field.name: value.amount, get_currency_field_name(field.name, field): value.currency}
        elif isinstance(value, F):
            # amount__in=[Money(1, 'EUR'), F('another_money')]
            target_field = _get_field(model, value.name)
            option = {
                field.name: value,
                get_currency_field_name(field.name, field): F(get_currency_field_name(value.name, target_field)),
            }
        else:
            # amount__in=[1, 2, 3]
            option = {field.name: value}
        new_query |= Q(**option)
    return new_query


def _expand_money_args(model, args):
    """
    Augments args so that they contain _currency lookups - ie.. Q() | Q()
    """
    for arg in args:
        if isinstance(arg, Q):
            _expand_arg(model, arg)
    return args


def _expand_arg(model, arg):
    for i, child in enumerate(arg.children):
        if isinstance(child, Q):
            _expand_arg(model, child)
        elif isinstance(child, (list, tuple)):
            name, value = child
            field = _get_field(model, name)
            if isinstance(value, MONEY_CLASSES):
                clean_name = _get_clean_name(model, name)
                currency_field_name = get_currency_field_name(clean_name, field)
                arg.children[i] = Q(child, (currency_field_name, smart_str(value.currency)))
            if isinstance(field, MoneyField):
                if isinstance(value, (BaseExpression, F)):
                    clean_name = _get_clean_name(model, name)
                    if not isinstance(value, F):
                        value = prepare_expression(value)
                    if not _is_money_field(model, value, name):
                        continue
                    currency_field_name = get_currency_field_name(clean_name, field)
                    target_field = _get_field(model, value.name)
                    arg.children[i] = Q(
                        child, (currency_field_name, F(get_currency_field_name(value.name, target_field)))
                    )
                if is_in_lookup(name, value):
                    arg.children[i] = _convert_in_lookup(model, name, value)


def _is_money_field(model, rhs, lhs_name):
    """
    Checks if the target field from the expression is instance of MoneyField.
    """
    # If the right side is the same field, then no reason to check
    if rhs.name == lhs_name:
        return True
    target_field = _get_field(model, rhs.name)
    return isinstance(target_field, MoneyField)


def _expand_money_kwargs(model, args=(), kwargs=None, exclusions=()):
    """
    Augments kwargs so that they contain _currency lookups.
    """
    for name, value in list(kwargs.items()):
        if name in exclusions:
            continue
        field = _get_field(model, name)
        if isinstance(value, MONEY_CLASSES):
            clean_name = _get_clean_name(model, name)
            kwargs[name] = value.amount
            currency_field_name = get_currency_field_name(clean_name, field)
            kwargs[currency_field_name] = smart_str(value.currency)
        else:
            if isinstance(field, MoneyField):
                if isinstance(value, (BaseExpression, F)) and not isinstance(value, (Case, Cast, Coalesce)):
                    clean_name = _get_clean_name(model, name)
                    if not isinstance(value, F):
                        value = prepare_expression(value)
                    if not _is_money_field(model, value, name):
                        continue
                    currency_field_name = get_currency_field_name(clean_name, field)
                    target_field = _get_field(model, value.name)
                    kwargs[currency_field_name] = F(get_currency_field_name(value.name, target_field))
                if is_in_lookup(name, value):
                    args += (_convert_in_lookup(model, name, value),)
                    del kwargs[name]
            elif isinstance(field, CurrencyField) and "defaults" in exclusions:
                _handle_currency_field(model, name, kwargs)

    return args, kwargs


def _handle_currency_field(model, name, kwargs):
    name = _get_clean_name(model, name)
    field = _get_field(model, name)
    money_field = field.price_field
    if money_field.default is not NOT_PROVIDED and money_field.name not in kwargs:
        kwargs["defaults"] = kwargs.get("defaults", {})
        kwargs["defaults"][money_field.name] = money_field.default.amount


def _get_model(args, func):
    """
    Returns the model class for given function.
    Note, that ``self`` is not available for proxy models.
    """
    if hasattr(func, "__self__"):
        # Bound method
        model = func.__self__.model
    elif hasattr(func, "__wrapped__"):
        # Proxy model
        model = func.__wrapped__.__self__.model
    else:
        # Custom method on user-defined model manager.
        model = args[0].model
    return model


def understands_money(func):
    """
    Used to wrap a queryset method with logic to expand
    a query from something like:

    mymodel.objects.filter(money=Money(100, "USD"))

    To something equivalent to:

    mymodel.objects.filter(money=Decimal("100.0"), money_currency="USD")
    """

    @wraps(func)
    def wrapper(*args, **kwargs):
        model = _get_model(args, func)
        args = _expand_money_args(model, args)
        exclusions = EXPAND_EXCLUSIONS.get(func.__name__, ())
        args, kwargs = _expand_money_kwargs(model, args, kwargs, exclusions)
        queryset = func(*args, **kwargs)
        return add_money_comprehension_to_queryset(queryset)

    return wrapper


RELEVANT_QUERYSET_METHODS = ("distinct", "get", "get_or_create", "filter", "exclude", "update", "order_by")
EXPAND_EXCLUSIONS = {"get_or_create": ("defaults",)}


def add_money_comprehension_to_queryset(qs):
    # Decorate each relevant method with understands_money in the queryset given
    for attr in RELEVANT_QUERYSET_METHODS:
        method = getattr(qs, attr, None)
        if method is not None:
            setattr(qs, attr, understands_money(method))
    return qs


def money_manager(manager):
    """
    Patches a model manager's get_queryset method so that each QuerySet it returns
    is able to work on money fields.

    This allow users of django-money to use other managers while still doing
    money queries.
    """

    # Need to dynamically subclass to add our behaviour, and then change
    # the class of 'manager' to our subclass.

    # Rejected alternatives:
    #
    # * A monkey patch that adds things to the manager instance dictionary.
    #   This fails due to complications with Manager._copy_to_model behaviour.
    #
    # * Returning a new MoneyManager instance (rather than modifying
    #   the passed in manager instance). This fails for reasons that
    #   are tricky to get to the bottom of - Manager does funny things.
    class MoneyManager(manager.__class__):
        def get_queryset(self, *args, **kwargs):
            queryset = super().get_queryset(*args, **kwargs)
            return add_money_comprehension_to_queryset(queryset)

    manager.__class__ = MoneyManager
    return manager



from django.conf import settings

import pytest

from djmoney.money import Money

from .testapp.models import RevisionedModel


@pytest.mark.django_db
@pytest.mark.skipif("reversion" not in settings.INSTALLED_APPS, reason="django-reversion not enabled")
def test_that_can_safely_restore_deleted_object():
    from reversion.models import Version
    from reversion.revisions import create_revision

    amount = Money(100, "GHS")
    with create_revision():
        instance = RevisionedModel.objects.create(amount=amount)
    instance.delete()
    version = Version.objects.get_deleted(RevisionedModel)[0]
    version.revision.revert()
    instance = RevisionedModel.objects.get(pk=1)
    assert instance.amount == amount



from django.utils.module_loading import import_string

from ..base import BaseExchangeCommand


class Command(BaseExchangeCommand):
    help = "Updates exchange rates."

    def handle(self, *args, **options):
        backend = import_string(options["backend"])()
        backend.update_rates()
        self.success("Successfully updated rates from %s" % backend.name)



from importlib import reload as reload_module

from django.db import models

import djmoney.settings
from djmoney.models.fields import MoneyField


class TestCurrencies:
    def create_class(self, **field_kwargs):
        class SettingsModel(models.Model):
            field = MoneyField(**field_kwargs)

            class Meta:
                app_label = "test"

        return SettingsModel

    def assert_choices(self, expected):
        reload_module(djmoney.settings)
        assert djmoney.settings.CURRENCY_CHOICES == expected

    def test_project_currencies(self, settings):
        settings.CURRENCIES = ["USD", "EUR"]
        self.assert_choices([("EUR", "Euro"), ("USD", "US Dollar")])

    def test_custom_currencies(self, settings):
        settings.CURRENCIES = ["USD", "EUR"]
        settings.CURRENCY_CHOICES = [("USD", "USD $"), ("EUR", "EUR €")]
        self.assert_choices([("EUR", "EUR €"), ("USD", "USD $")])

    def test_default_currency(self, settings):
        settings.DEFAULT_CURRENCY = None
        klass = self.create_class(default_currency=None, max_digits=10, decimal_places=2, null=True)
        assert klass._meta.fields[2].default_currency is None
        instance = klass()
        assert instance.field is None



# Generated by Django 2.0.4 on 2018-04-03 08:42

import django.db.models.deletion
from django.db import migrations, models

from djmoney.settings import CURRENCY_CODE_MAX_LENGTH


class Migration(migrations.Migration):

    initial = True

    dependencies = []

    operations = [
        migrations.CreateModel(
            name="ExchangeBackend",
            fields=[
                ("name", models.CharField(max_length=255, primary_key=True, serialize=False)),
                ("last_update", models.DateTimeField(auto_now=True)),
                ("base_currency", models.CharField(max_length=CURRENCY_CODE_MAX_LENGTH)),
            ],
        ),
        migrations.CreateModel(
            name="Rate",
            fields=[
                ("id", models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                ("currency", models.CharField(max_length=CURRENCY_CODE_MAX_LENGTH)),
                ("value", models.DecimalField(decimal_places=6, max_digits=20)),
                (
                    "backend",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE, to="exchange.ExchangeBackend", related_name="rates"
                    ),
                ),
            ],
        ),
        migrations.AlterUniqueTogether(name="rate", unique_together={("currency", "backend")}),
    ]



import json
from unittest.mock import patch

import django
from django.core.management import call_command
from django.core.serializers.base import DeserializationError

import pytest

from djmoney.money import Money
from djmoney.serializers import Deserializer, Serializer

from .testapp.models import ModelWithDefaultAsInt, ModelWithSharedCurrency


pytestmark = pytest.mark.django_db


def test_m2m_fields_are_not_lost(concrete_instance, m2m_object):
    """
    See #184.
    """
    value = Serializer().serialize([concrete_instance])
    m2m_data = list(Deserializer(value, ignorenonexistent=True))[0].m2m_data
    assert m2m_object in ModelWithDefaultAsInt.objects.filter(pk__in=m2m_data["m2m_field"])


@pytest.fixture
def fixture_file(tmpdir):
    return tmpdir.join("dump.json")


def dumpdata(capsys):
    call_command("dumpdata", "testapp")
    return capsys.readouterr()[0]


def loaddata(fixture_file, ignore_value=False):
    call_command("loaddata", str(fixture_file), ignorenonexistent=ignore_value)


def test_dumpdata(capsys, fixture_file):
    money = Money(10, "EUR")
    instance = ModelWithDefaultAsInt.objects.create(money=money)
    data = dumpdata(capsys)
    fixture_file.write(data)
    instance.delete()
    loaddata(fixture_file)
    assert ModelWithDefaultAsInt.objects.get().money == money


def test_load_invalid(fixture_file):
    data = '[{"model": "testapp.unknown_model", "pk": 1, "fields": {"money_currency": "USD", "money": "1.00"}}]'
    fixture_file.write(data)
    with pytest.raises(DeserializationError):
        loaddata(fixture_file)


def test_load_invalid_ignore(fixture_file):
    data = (
        '[{"model": "testapp.unknown_model", "pk": 1, "fields": {"money_currency": "USD", "money": "1.00"}}, '
        '{"model": "testapp.modelwithdefaultasint", "pk": 2, "fields": {"money_currency": "USD", "money": "1.00"}}]'
    )
    fixture_file.write(data)
    loaddata(fixture_file, True)


def test_old_fields_skip(capsys, fixture_file):
    money = Money(10, "EUR")
    instance = ModelWithDefaultAsInt.objects.create(money=money)
    data = dumpdata(capsys)
    instance.delete()
    out = json.loads(data)
    out[0]["fields"]["old_field"] = 1
    out = json.dumps(out)
    fixture_file.write(out)
    loaddata(fixture_file, True)
    assert ModelWithDefaultAsInt.objects.get().money == money


def test_deserialization_error():
    with pytest.raises(DeserializationError):
        list(Deserializer("invalid JSON"))


@pytest.mark.skipif(django.VERSION >= (5, 2), reason="_get_model was removed in Django 5.2")
def test_patched_get_model(fixture_file):
    """Sometimes it is useful to patch `django.code.serializers.python._get_model`.

    Our code should use the patched version."""
    data = '[{"model": "testapp.unknown_model", "pk": 1, "fields": {"money_currency": "USD", "money": "1.00"}}]'
    fixture_file.write(data)

    def _get_model(identifier):
        return ModelWithDefaultAsInt

    with patch("django.core.serializers.python._get_model", _get_model):
        loaddata(fixture_file)
    assert ModelWithDefaultAsInt.objects.get().money == Money(1, "USD")


@pytest.mark.skipif(django.VERSION < (5, 2), reason="Deserializer API were added in Django 5.2")
def test_patched_get_model_from_node(fixture_file):
    """Sometimes it is useful to patch `django.code.serializers.python.Deserializer._get_model_from_node`.

    Our code should use the patched version."""
    data = '[{"model": "testapp.unknown_model", "pk": 1, "fields": {"money_currency": "USD", "money": "1.00"}}]'
    fixture_file.write(data)

    def _get_model_from_node(self, identifier):
        return ModelWithDefaultAsInt

    with patch("django.core.serializers.python.Deserializer._get_model_from_node", _get_model_from_node):
        loaddata(fixture_file)
    assert ModelWithDefaultAsInt.objects.get().money == Money(1, "USD")


def test_serialize_currency_field(fixture_file):
    data = """[
    {
        "model": "testapp.modelwithsharedcurrency",
        "pk": 1,
        "fields": {"first": "1.00", "second": "2.00", "currency": "USD"}
    }
]"""
    fixture_file.write(data)
    loaddata(fixture_file)
    assert ModelWithSharedCurrency.objects.get().first == Money(1, "USD")



"""
Created on May 15, 2011

@author: jake
"""

from django import forms

from djmoney.forms import MoneyField

from .models import (
    ModelWithCallableCurrencyChoices,
    ModelWithCallableDefault,
    ModelWithCallableDefaultAndDefaultCurrency,
    ModelWithCallableDefaultCurrency,
    ModelWithDefaultAsString,
    ModelWithDefaultPrecision,
    ModelWithValidation,
    ModelWithVanillaMoneyField,
    NullMoneyFieldModel,
    PositiveValidatedMoneyModel,
    PreciseModel,
    SimpleModel,
    ValidatedMoneyModel,
)


class MoneyForm(forms.Form):
    money = MoneyField(currency_choices=[("SEK", "Swedish Krona")], max_value=1000, min_value=2)


class MoneyFormMultipleCurrencies(forms.Form):
    money = MoneyField(currency_choices=[("SEK", "Swedish Krona"), ("EUR", "Euro")], max_value=1000, min_value=2)


class OptionalMoneyForm(forms.Form):
    money = MoneyField(required=False, currency_choices=[("SEK", "Swedish Krona")])


class MoneyModelForm(forms.ModelForm):
    class Meta:
        model = ModelWithVanillaMoneyField
        fields = ("money",)


class MoneyModelFormWithCallableDefault(forms.ModelForm):
    class Meta:
        model = ModelWithCallableDefault
        fields = ("money",)


class MoneyModelFormWithCallableDefaultCurrency(forms.ModelForm):
    class Meta:
        model = ModelWithCallableDefaultCurrency
        fields = ("money",)


class MoneyModelFormWithCallableCurrencyChoices(forms.ModelForm):
    class Meta:
        model = ModelWithCallableCurrencyChoices
        fields = ("money",)


class MoneyModelFormWithCallableDefaultAndCallableDefaultCurrency(forms.ModelForm):
    class Meta:
        model = ModelWithCallableDefaultAndDefaultCurrency
        fields = ("money",)


class NullableModelForm(forms.ModelForm):
    class Meta:
        model = NullMoneyFieldModel
        fields = ("field",)


class DefaultMoneyModelForm(forms.ModelForm):
    class Meta:
        model = ModelWithDefaultAsString
        fields = ("money",)


class DefaultPrecisionModelForm(forms.ModelForm):
    class Meta:
        model = ModelWithDefaultPrecision
        fields = ("money",)


class MoneyModelFormWithValidation(forms.ModelForm):
    class Meta:
        model = ModelWithValidation
        fields = ("balance",)


class ValidatedMoneyModelForm(forms.ModelForm):
    class Meta:
        model = ValidatedMoneyModel
        fields = ("money",)


class PositiveValidatedMoneyModelForm(forms.ModelForm):
    class Meta:
        model = PositiveValidatedMoneyModel
        fields = ("money",)


class DisabledFieldForm(forms.ModelForm):
    money = MoneyField(disabled=True)

    class Meta:
        fields = "__all__"
        model = SimpleModel


class PreciseModelForm(forms.ModelForm):
    class Meta:
        fields = "__all__"
        model = PreciseModel


class PreciseForm(forms.Form):
    money = MoneyField()

    class Meta:
        fields = "__all__"



from decimal import Decimal

from django.core.exceptions import ValidationError
from django.core.validators import DecimalValidator
from django.db import models
from django.db.models import NOT_PROVIDED, F, Field, Func, Value
from django.db.models.expressions import BaseExpression
from django.db.models.signals import class_prepared
from django.utils.encoding import smart_str
from django.utils.functional import cached_property

from djmoney import forms
from djmoney.money import Currency, Money
from moneyed import Money as OldMoney

from .._compat import setup_managers
from ..settings import CURRENCY_CHOICES, CURRENCY_CODE_MAX_LENGTH, DECIMAL_PLACES, DEFAULT_CURRENCY
from ..utils import MONEY_CLASSES, get_currency_field_name, prepare_expression


__all__ = ("MoneyField",)


class MoneyValidator(DecimalValidator):
    def __call__(self, value):
        return super().__call__(value.amount)


def get_value(obj, expr):
    """
    Extracts value from object or expression.
    """
    if isinstance(expr, F):
        expr = getattr(obj, expr.name)
    else:
        expr = expr.value
    if isinstance(expr, OldMoney):
        expr = Money(expr.amount, expr.currency)
    return expr


def validate_money_expression(obj, expr):
    """
    Money supports different types of expressions, but you can't do following:
      - Add or subtract money with not-money
      - Any exponentiation
      - Any operations with money in different currencies
      - Multiplication, division, modulo with money instances on both sides of expression
    """
    connector = expr.connector
    lhs = get_value(obj, expr.lhs)
    rhs = get_value(obj, expr.rhs)

    if (not isinstance(rhs, Money) and connector in ("+", "-")) or connector == "^":
        raise ValidationError("Invalid F expression for MoneyField.", code="invalid")
    if isinstance(lhs, Money) and isinstance(rhs, Money):
        if connector in ("*", "/", "^", "%%"):
            raise ValidationError("Invalid F expression for MoneyField.", code="invalid")
        if lhs.currency != rhs.currency:
            raise ValidationError("You cannot use F() with different currencies.", code="invalid")


def validate_money_value(value):
    """
    Valid value for money are:
      - Single numeric value
      - Money instances
      - Pairs of numeric value and currency. Currency can't be None.
    """
    if isinstance(value, (list, tuple)) and (len(value) != 2 or value[1] is None):
        raise ValidationError("Invalid value for MoneyField: %(value)s.", code="invalid", params={"value": value})


def get_currency(value):
    """
    Extracts currency from value.
    """
    if isinstance(value, MONEY_CLASSES):
        return smart_str(value.currency)
    elif isinstance(value, (list, tuple)):
        return value[1]


class MoneyFieldProxy:
    def __init__(self, field):
        self.field = field
        self.currency_field_name = get_currency_field_name(self.field.name, self.field)

    def _money_from_obj(self, obj):
        amount = obj.__dict__[self.field.name]
        currency = obj.__dict__[self.currency_field_name]
        if amount is None:
            return None
        elif currency is None:
            raise TypeError("Currency code can't be None")
        return Money(amount=amount, currency=currency, decimal_places=self.field.decimal_places)

    def __get__(self, obj, type=None):
        if obj is None:
            return self
        data = obj.__dict__
        if isinstance(data[self.field.name], BaseExpression):
            return data[self.field.name]
        if not isinstance(data[self.field.name], Money):
            data[self.field.name] = self._money_from_obj(obj)
        return data[self.field.name]

    def __set__(self, obj, value):  # noqa
        if (
            value is not None
            and self.field._currency_field.null
            and not isinstance(value, MONEY_CLASSES)
            and not obj.__dict__[self.currency_field_name]
        ):
            # For nullable fields we need either both NULL amount and currency or both NOT NULL
            raise ValueError("Missing currency value")
        if isinstance(value, BaseExpression):
            if isinstance(value, Value):
                value = self.prepare_value(obj, value.value)
            elif not isinstance(value, Func):
                validate_money_expression(obj, value)
                prepare_expression(value)
        else:
            value = self.prepare_value(obj, value)
        obj.__dict__[self.field.name] = value

    def prepare_value(self, obj, value):
        validate_money_value(value)
        currency = get_currency(value)
        if currency:
            self.set_currency(obj, currency)
        return self.field.to_python(value)

    def set_currency(self, obj, value):
        # we have to determine whether to replace the currency.
        # i.e. if we do the following:
        # .objects.get_or_create(money_currency='EUR')
        # then the currency is already set up, before this code hits
        # __set__ of MoneyField. This is because the currency field
        # has less creation counter than money field.
        #
        # Gotcha:
        # But we should also allow setting a field back to its original default
        # value!
        # https://github.com/django-money/django-money/issues/221
        object_currency = obj.__dict__.get(self.currency_field_name)
        if object_currency != value:
            # in other words, update the currency only if it wasn't
            # changed before.
            setattr(obj, self.currency_field_name, value)


class CurrencyField(models.CharField):
    description = "A field which stores currency."

    def __init__(self, price_field=None, default=DEFAULT_CURRENCY, **kwargs):
        if isinstance(default, Currency):
            default = default.code
        kwargs.setdefault("max_length", CURRENCY_CODE_MAX_LENGTH)
        self.price_field = price_field
        super().__init__(default=default, **kwargs)

    def contribute_to_class(self, cls, name):
        if name not in [f.name for f in cls._meta.fields]:
            super().contribute_to_class(cls, name)


class MoneyField(models.DecimalField):
    description = "A field which stores both the currency and amount of money."

    @property
    def non_db_attrs(self):
        return super().non_db_attrs + ("default_currency", "currency_choices")

    def __init__(
        self,
        verbose_name=None,
        name=None,
        max_digits=None,
        decimal_places=DECIMAL_PLACES,
        default=NOT_PROVIDED,
        default_currency=DEFAULT_CURRENCY,
        currency_choices=CURRENCY_CHOICES,
        currency_max_length=CURRENCY_CODE_MAX_LENGTH,
        currency_field_name=None,
        money_descriptor_class=MoneyFieldProxy,
        **kwargs,
    ):
        nullable = kwargs.get("null", False)
        default = self.setup_default(default, default_currency, nullable)

        # Provide the default currency from the default Money instance provided
        if not default_currency and isinstance(default, Money):
            default_currency = default.currency

        self.currency_max_length = currency_max_length
        self.default_currency = default_currency
        self.currency_choices = currency_choices
        self.currency_field_name = currency_field_name
        self.money_descriptor_class = money_descriptor_class

        super().__init__(verbose_name, name, max_digits, decimal_places, default=default, **kwargs)
        self.creation_counter += 1
        Field.creation_counter += 1

    def setup_default(self, default, default_currency, nullable):
        # None and NOT_PROVIDED are returned as-is
        if default in (None, NOT_PROVIDED):
            return default
        # Money instances need no further processing
        elif isinstance(default, Money):
            return default
        # Callables continue to be callables (will be evaluated later)
        # Evaluating them here will create new migrations every time the callable returns a new value
        elif callable(default):
            return default
        # This is not supported: If default_currency is a callable, we need to fail, we cannot call it to construct the
        # default, that would have unintended consequences. Require to have BOTH default and default_currency as
        # callables in that case.
        elif callable(default_currency) and default_currency != NOT_PROVIDED:
            raise ValueError(
                "django-money does not support a callable for default_currency, if the amount's default isn't also a "
                "callable."
            )
        elif isinstance(default, (str, bytes)):
            try:
                # handle scenario where default is formatted like:
                # 'amount currency-code'
                amount, currency = (default.decode() if isinstance(default, bytes) else default).split(" ", 1)
            except ValueError:
                # value error would be risen if the default is
                # without the currency part, i.e
                # 'amount'
                amount = default.decode() if isinstance(default, bytes) else default
                currency = default_currency

            amount = Decimal(amount)
        elif isinstance(default, (float, Decimal, int)):
            amount, currency = default, default_currency
        elif isinstance(default, OldMoney):
            amount, currency = default.amount, default.currency
        else:
            raise ValueError(f"default value must be an instance of Money, is: {default}")

        assert currency is not None, (
            "Default currency can not be `None` when default is of a value that doesn't include a currency. Either"
            " provide a default containing a currency value or configure a currency default via setting"
            " `default_currency=` or `DEFAULT_CURRENCY`(global)"
        )
        return Money(amount, currency)

    def to_python(self, value):
        if isinstance(value, MONEY_CLASSES):
            value = value.amount
        elif isinstance(value, tuple):
            value = value[0]
        if isinstance(value, float):
            value = str(value)
        return super().to_python(value)

    def clean(self, value, model_instance):
        """
        We need to run validation against ``Money`` instance.
        """
        output = self.to_python(value)
        self.validate(value, model_instance)
        self.run_validators(value)
        return output

    @cached_property
    def validators(self):
        """
        Default ``DecimalValidator`` doesn't work with ``Money`` instances.
        """
        return super(models.DecimalField, self).validators + [MoneyValidator(self.max_digits, self.decimal_places)]

    def contribute_to_class(self, cls, name):
        cls._meta.has_money_field = True

        # Note the discussion about whether or not the currency field should be added in migrations:
        # https://github.com/django-money/django-money/issues/725
        # https://github.com/django-money/django-money/pull/726
        # https://github.com/django-money/django-money/issues/731
        if not hasattr(self, "_currency_field"):
            self.add_currency_field(cls, name)

        super().contribute_to_class(cls, name)

        setattr(cls, self.name, self.money_descriptor_class(self))

    def add_currency_field(self, cls, name):
        """
        Adds CurrencyField instance to a model class.
        """
        currency_field = CurrencyField(
            price_field=self,
            max_length=self.currency_max_length,
            default=self.default_currency,
            editable=False,
            choices=self.currency_choices,
            null=self.null,
        )
        currency_field.creation_counter = self.creation_counter - 1
        currency_field_name = get_currency_field_name(name, self)
        cls.add_to_class(currency_field_name, currency_field)
        self._currency_field = currency_field

    def get_db_prep_save(self, value, connection):
        if isinstance(value, MONEY_CLASSES):
            value = value.amount
        return super().get_db_prep_save(value, connection)

    def get_default(self):
        if isinstance(self.default, Money):
            return self.default
        else:
            return super().get_default()

    @property
    def _has_default(self):
        # Whether the field has an explicitly provided non-empty default.
        # `None` was used by django-money before, and we need to check it because it can come from historical migrations
        return self.default not in (None, NOT_PROVIDED)

    def formfield(self, **kwargs):
        defaults = {"form_class": forms.MoneyField, "decimal_places": self.decimal_places}
        defaults.update(kwargs)
        defaults["currency_choices"] = self.currency_choices
        defaults["default_currency"] = self.default_currency
        if self._has_default:
            if callable(self.default):
                defaults["default_amount"] = self.default().amount
            else:
                defaults["default_amount"] = self.default.amount
        return super().formfield(**defaults)

    def value_to_string(self, obj):
        value = self.value_from_object(obj)
        return self.get_prep_value(value)

    def deconstruct(self):
        name, path, args, kwargs = super().deconstruct()

        # Comment needed: Why is this done?
        if self.default_currency != DEFAULT_CURRENCY:
            kwargs["default_currency"] = self.default_currency
        # Comment needed: Why is this done?
        if self.currency_choices != CURRENCY_CHOICES:
            kwargs["currency_choices"] = self.currency_choices
        if self.currency_field_name:
            kwargs["currency_field_name"] = self.currency_field_name
        if self.currency_max_length != CURRENCY_CODE_MAX_LENGTH:
            kwargs["currency_max_length"] = self.currency_max_length
        return name, path, args, kwargs


def patch_managers(sender, **kwargs):
    """
    Patches models managers.
    """
    if sender._meta.proxy_for_model:
        has_money_field = hasattr(sender._meta.proxy_for_model._meta, "has_money_field")
    else:
        has_money_field = hasattr(sender._meta, "has_money_field")

    if has_money_field:
        setup_managers(sender)


class_prepared.connect(patch_managers)



from django.contrib.admin import site

import pytest

from djmoney.contrib.exchange.admin import RateAdmin
from djmoney.contrib.exchange.models import Rate


pytestmark = pytest.mark.django_db


def test_last_update(backend):
    rate = Rate(currency="NOK", value=5, backend=backend)
    assert RateAdmin(Rate, site).last_update(rate) == backend.last_update



from django.core.exceptions import ValidationError
from django.forms import ChoiceField, DecimalField, MultiValueField

from djmoney.money import Money
from djmoney.utils import MONEY_CLASSES

from ..settings import CURRENCY_CHOICES, DECIMAL_PLACES
from .widgets import HiddenMoneyWidget, MoneyWidget


__all__ = ("MoneyField",)


class MoneyField(MultiValueField):
    def __init__(
        self,
        currency_widget=None,
        currency_choices=CURRENCY_CHOICES,
        max_value=None,
        min_value=None,
        max_digits=None,
        decimal_places=DECIMAL_PLACES,
        default_amount=None,
        default_currency=None,
        *args,
        **kwargs,
    ):

        amount_field = DecimalField(
            *args,
            max_value=max_value,
            min_value=min_value,
            max_digits=max_digits,
            decimal_places=decimal_places,
            **kwargs,
        )
        currency_field = ChoiceField(choices=currency_choices)

        self.widget = MoneyWidget(
            amount_widget=amount_field.widget,
            currency_widget=currency_widget or currency_field.widget,
            default_currency=default_currency,
        )
        # The two fields that this widget comprises
        fields = (amount_field, currency_field)

        super().__init__(fields, *args, **kwargs)

        # set the initial value to the default currency so that the
        # default currency appears as the selected menu item

        # Callables might be supplied in cases where the initial data is directly copied from the MoneyField db
        # object - this seems to happen internally when Django auto-generates certain hidden input fields to track
        # initial data in a formset.
        if callable(default_currency):
            default_currency = default_currency()

        # TODO: are we sure we do not have default_amount callables here?
        if isinstance(default_amount, Money):
            default_amount = default_amount.amount

        self.initial = [default_amount, default_currency]

    def hidden_widget(self):
        # TODO: This should inherit the constraints of the field
        #  Otherwise, we won't validate that min, max value and currency_choices are valid?
        # This is usually used to pre-fill the 'initial data' hidden input, so it's not very sensitive to tampering.
        return HiddenMoneyWidget()

    def compress(self, data_list):
        if data_list:
            if not self.required and data_list[0] in self.empty_values:
                return None
            else:
                return Money(*data_list[:2])
        return None

    def clean(self, value):
        if isinstance(value, MONEY_CLASSES):
            value = (value.amount, value.currency)
        # Unpack the list and fail if it doesn't match the expected structure
        if isinstance(value, (list, tuple)):
            amount, currency = value
            value = (amount, currency)
        return super().clean(value)

    def has_changed(self, initial, data):  # noqa
        if self.disabled:
            return False
        if initial is None:
            initial = ["" for _ in range(0, len(data))]
        else:
            # If the initial value was supplied as a list, go with that.
            # Otherwise, try to decompress it.
            if not isinstance(initial, list):
                initial = self.widget.decompress(initial)

        amount_field, currency_field = self.fields
        amount_initial, currency_initial = initial

        # We treat the amount and currency fields slightly
        # differently: if the amount has changed, then we definitely
        # consider the money value to have changed. If the currency
        # has changed, but the amount is *empty* then we do not
        # consider the money value to have changed. This means that it
        # plays nicely with empty form rows in formsets.
        try:
            amount_data = data[0]
        except IndexError:
            amount_data = None

        try:
            amount_initial = amount_field.to_python(amount_initial)
        except ValidationError:
            return True
        if amount_field.has_changed(amount_initial, amount_data):
            return True

        try:
            currency_data = data[1]
        except IndexError:
            currency_data = None

        try:
            currency_initial = currency_field.to_python(currency_initial)
        except ValidationError:
            return True
        # If the currency is valid, has changed and there is some
        # amount data, then the money value has changed.
        if currency_field.has_changed(currency_initial, currency_data) and amount_data:
            return True

        return False



from collections import Counter
from decimal import Decimal

from django.test import override_settings

import pytest

from djmoney.money import Money

from ..testapp.models import (
    ModelWithVanillaMoneyField,
    MoneyFieldModelWithProperty,
    NullMoneyFieldModel,
    ValidatedMoneyModel,
)


pytestmark = pytest.mark.django_db
serializers = pytest.importorskip("rest_framework.serializers")
fields = pytest.importorskip("rest_framework.fields")
djmoney_fields = pytest.importorskip("djmoney.contrib.django_rest_framework.fields")


class TestMoneyField:
    def get_serializer(
        self, model_class, field_name=None, instance=None, data=fields.empty, fields_="__all__", field_kwargs=None
    ):
        class MetaSerializer(serializers.SerializerMetaclass):
            def __new__(cls, name, bases, attrs):
                from djmoney.contrib.django_rest_framework import MoneyField

                if field_name is not None and field_kwargs is not None:
                    attrs[field_name] = MoneyField(max_digits=10, decimal_places=2, **field_kwargs)
                return super().__new__(cls, name, bases, attrs)

        class Serializer(serializers.ModelSerializer, metaclass=MetaSerializer):
            class Meta:
                model = model_class
                fields = fields_

        return Serializer(instance=instance, data=data)

    @pytest.mark.parametrize(
        "model_class, create_kwargs, expected",
        (
            (NullMoneyFieldModel, {"field": None}, {"field": None, "field_currency": "USD"}),
            (NullMoneyFieldModel, {"field": Money(10, "USD")}, {"field": "10.00", "field_currency": "USD"}),
            (
                ModelWithVanillaMoneyField,
                {"money": Money(10, "USD")},
                {
                    "integer": 0,
                    "money": "10.00",
                    "money_currency": "USD",
                    "second_money": "0.00",
                    "second_money_currency": "EUR",
                },
            ),
        ),
    )
    def test_to_representation(self, model_class, create_kwargs, expected):
        instance = model_class.objects.create(**create_kwargs)
        expected["id"] = instance.id
        serializer = self.get_serializer(model_class, instance=instance)
        assert serializer.data == expected

    @pytest.mark.parametrize(
        "model_class, field, field_kwargs, value, expected",
        (
            (NullMoneyFieldModel, "field", None, None, None),
            (NullMoneyFieldModel, "field", {"default_currency": "EUR", "allow_null": True}, None, None),
            (NullMoneyFieldModel, "field", None, Money(10, "USD"), Money(10, "USD")),
            (NullMoneyFieldModel, "field", {"default_currency": "EUR"}, Money(10, "USD"), Money(10, "USD")),
            (ModelWithVanillaMoneyField, "money", {"default_currency": "EUR"}, 10, Money(10, "EUR")),
            (ModelWithVanillaMoneyField, "money", None, Money(10, "USD"), Money(10, "USD")),
            (ModelWithVanillaMoneyField, "money", {"default_currency": "EUR"}, Money(10, "USD"), Money(10, "USD")),
            (ModelWithVanillaMoneyField, "money", {"default_currency": "EUR"}, 10, Money(10, "EUR")),
        ),
    )
    def test_to_internal_value(self, model_class, field, field_kwargs, value, expected):
        serializer = self.get_serializer(model_class, field_name=field, data={field: value}, field_kwargs=field_kwargs)
        assert serializer.is_valid()
        instance = serializer.save()
        assert getattr(instance, field) == expected

    def test_to_internal_value_for_property_field(self):

        class PropertyModelSerializer(serializers.ModelSerializer):
            extra_monies = djmoney_fields.MoneyField(
                source="ten_extra_monies",
                max_digits=10,
                decimal_places=2,
                min_value=0,
                default_currency="EUR",
                required=False,
                allow_null=True,
                read_only=True,
            )

            class Meta:
                model = MoneyFieldModelWithProperty
                fields = ("extra_monies", "money")

        serializer = PropertyModelSerializer(data={"money": Money(12, "USD"), "ten_extra_monies": Money(100, "USD")})
        serializer.is_valid()
        assert serializer.validated_data == {"money": Money("12.00", "USD")}
        instance = serializer.save()
        assert instance.ten_extra_monies == Money(22, "USD")
        assert instance.money == Money(12, "USD")

    def test_invalid_value(self):
        serializer = self.get_serializer(ModelWithVanillaMoneyField, data={"money": None})
        assert not serializer.is_valid()
        error_text = "This field may not be null."
        assert serializer.errors == {"money": [error_text]}

    @pytest.mark.parametrize(
        "body, field_kwargs, expected",
        (
            ({"field": "10", "field_currency": "EUR"}, None, Money(10, "EUR")),
            ({"field": "10"}, {"default_currency": "EUR"}, Money(10, "EUR")),
            ({"field": "12.20", "field_currency": "GBP"}, None, Money(12.20, "GBP")),
            ({"field": "15.15", "field_currency": "USD"}, None, Money(15.15, "USD")),
            ({"field": None, "field_currency": None}, None, None),
            ({"field": None, "field_currency": None}, {"default_currency": "EUR"}, None),
            ({"field": "16", "field_currency": None}, None, Decimal("16.00")),
            ({"field": "16", "field_currency": None}, {"default_currency": "EUR"}, Decimal("16.00")),
            ({"field": None, "field_currency": "USD"}, None, None),
            ({"field": None, "field_currency": "USD"}, {"default_currency": "EUR"}, None),
        ),
    )
    def test_post_put_values(self, body, field_kwargs, expected):
        if field_kwargs is not None:
            field_kwargs["allow_null"] = True
        serializer = self.get_serializer(NullMoneyFieldModel, data=body, field_name="field", field_kwargs=field_kwargs)
        serializer.is_valid()
        assert serializer.validated_data["field"] == expected

    def test_serializer_with_fields(self):
        serializer = self.get_serializer(ModelWithVanillaMoneyField, data={"money": "10.00"}, fields_=("money",))
        serializer.is_valid(raise_exception=True)
        assert serializer.data == {"money": "10.00"}

    @pytest.mark.parametrize(
        "value, error",
        (
            (Money(50, "EUR"), "Ensure this value is greater than or equal to €100.00."),
            (Money(1500, "EUR"), "Ensure this value is less than or equal to €1,000.00."),
            (Money(40, "USD"), "Ensure this value is greater than or equal to $50.00."),
            (Money(600, "USD"), "Ensure this value is less than or equal to $500.00."),
            (Money(400, "NOK"), "Ensure this value is greater than or equal to NOK500.00."),
            (Money(950, "NOK"), "Ensure this value is less than or equal to NOK900.00."),
            (Money(10, "DKK"), "Ensure this value is greater than or equal to DKK50.00."),
            (Money(2500, "DKK"), "Ensure this value is less than or equal to DKK2,000.00."),
            (Money(5, "SEK"), "Ensure this value is greater than or equal to 10."),
            (Money(1600, "SEK"), "Ensure this value is less than or equal to 1500."),
        ),
    )
    def test_model_validators(self, value, error):
        serializer = self.get_serializer(
            ValidatedMoneyModel, data={"money": value.amount, "money_currency": value.currency.code}
        )
        assert not serializer.is_valid()
        assert serializer.errors["money"][0] == error

    @pytest.mark.parametrize(
        "value, error",
        (
            (Money(50, "EUR"), "Ensure this value is greater than or equal to 100."),
            (Money(1500, "EUR"), "Ensure this value is less than or equal to 1000."),
        ),
    )
    def test_boundary_values(self, value, error):
        serializer = self.get_serializer(
            NullMoneyFieldModel,
            data={"field": value.amount, "field_currency": value.currency.code},
            field_name="field",
            field_kwargs={"min_value": 100, "max_value": 1000},
        )
        assert not serializer.is_valid()
        assert serializer.errors["field"][0] == error

    @pytest.mark.parametrize(
        ("data", "error_codes"),
        [
            pytest.param(
                {"money": "", "money_currency": "XUA"},
                [("money", "invalid")],
                id="amount_as_empty_string",
            ),
            pytest.param(
                {"money": None, "money_currency": "XUA"},
                [("money", "null")],
                id="amount_as_none",
            ),
            pytest.param(
                {"money": "v", "money_currency": "XUA"},
                [("money", "invalid")],
                id="amount_as_invalid_decimal",
            ),
            pytest.param(
                {"money": "0.01", "money_currency": "v"},
                [("money", "invalid_currency")],
                id="invalid_currency",
            ),
            pytest.param(
                {"money_currency": "SEK"},
                [("money", "required")],
                id="amount_key_not_in_data",
            ),
        ],
    )
    def test_errors_on(self, data, error_codes):
        class Serializer(serializers.Serializer):
            money = djmoney_fields.MoneyField(max_digits=9, decimal_places=2)

        serializer = Serializer(data=data)
        with pytest.raises(serializers.ValidationError) as err:
            serializer.is_valid(raise_exception=True)

        assert Counter([(field, code) for field, codes in err.value.get_codes().items() for code in codes]) == Counter(
            error_codes
        )

    @pytest.mark.parametrize(
        ("data", "expected"),
        [
            pytest.param({"money": "0.01", "money_currency": None}, Decimal("0.01"), id="is_none"),
            pytest.param({"money": "0.01", "money_currency": ""}, Decimal("0.01"), id="is_empty_string"),
            pytest.param({"money": "0.01"}, Decimal("0.01"), id="key_not_in_data"),
        ],
    )
    def test_returns_decimal_when_currency(self, data, expected):
        class Serializer(serializers.Serializer):
            money = djmoney_fields.MoneyField(max_digits=9, decimal_places=2)

        serializer = Serializer(data=data)
        serializer.is_valid(raise_exception=True)
        assert serializer.validated_data["money"] == expected


# Test case contributed for
# https://github.com/django-money/django-money/pull/722
class TestMinValueSerializer:

    @override_settings(DEFAULT_CURRENCY="EUR")
    @pytest.mark.parametrize(
        ("data", "is_valid"),
        [
            pytest.param({"money": Money(-1, "EUR")}, False, id="is_invalid_money_value"),
            pytest.param({"money": Money(1, "EUR")}, True, id="is_valid_money_value"),
            pytest.param({"money": "-1", "money_currency": "EUR"}, False, id="is_invalid_dict_value"),
            pytest.param({"money": "0.01", "money_currency": "EUR"}, True, id="is_valid_dict_value"),
        ],
    )
    def test_serializer_validator_field_without_default_currency(self, data, is_valid):
        from djmoney.contrib.django_rest_framework import MoneyField

        class MinValueSerializer(serializers.Serializer):
            money = MoneyField(decimal_places=2, max_digits=10, min_value=0)

            class Meta:
                model = ModelWithVanillaMoneyField

        serializer = MinValueSerializer(data=data)
        if is_valid:
            assert serializer.is_valid()
        else:
            assert not serializer.is_valid()
            assert serializer.errors["money"][0] == "Ensure this value is greater than or equal to 0."

    @pytest.mark.parametrize(
        ("data", "is_valid"),
        [
            pytest.param({"second_money": Money(-1, "EUR")}, False, id="is_invalid_money_value"),
            pytest.param({"second_money": Money(1, "EUR")}, True, id="is_valid_money_value"),
            pytest.param({"second_money": "-1", "second_money_currency": "EUR"}, False, id="is_invalid_dict_value"),
            pytest.param({"second_money": "0.01", "second_money_currency": "EUR"}, True, id="is_valid_dict_value"),
        ],
    )
    def test_serializer_validator_field_with_default_currencey(self, data, is_valid):
        from djmoney.contrib.django_rest_framework import MoneyField

        class MinValueSerializer(serializers.Serializer):
            second_money = MoneyField(decimal_places=2, max_digits=10, min_value=0)

            class Meta:
                model = ModelWithVanillaMoneyField

        serializer = MinValueSerializer(data=data)
        if is_valid:
            assert serializer.is_valid()
        else:
            assert not serializer.is_valid()
            assert serializer.errors["second_money"][0] == "Ensure this value is greater than or equal to 0."

    def test_no_model_serializer(self):
        from djmoney.contrib.django_rest_framework import MoneyField

        class NormalSerializer(serializers.Serializer):
            the_money = MoneyField(decimal_places=2, max_digits=10, min_value=0)

            class Meta:
                fields = ("the_money",)

        serializer = NormalSerializer(data={"the_money": "0.01", "the_money_currency": "EUR"})
        assert serializer.is_valid()

    def test_model_serializer_with_field_source(self):
        class ModelSerializer(serializers.ModelSerializer):
            renamed_money_field = djmoney_fields.MoneyField(
                source="money", max_digits=10, decimal_places=2, min_value=0
            )

            class Meta:
                model = ModelWithVanillaMoneyField
                fields = ("renamed_money_field",)

        serializer = ModelSerializer(data={"renamed_money_field": "0.01", "renamed_money_field_currency": "EUR"})
        assert serializer.is_valid()

    def test_model_serializer_with_nonexistent_property_raises_error(self):
        class PropertyModelSerializer(serializers.ModelSerializer):
            nonexistent_field = djmoney_fields.MoneyField(
                max_digits=10,
                decimal_places=2,
                min_value=0,
                default_currency="EUR",
                required=False,
                allow_null=True,
            )

            class Meta:
                model = MoneyFieldModelWithProperty
                fields = ("nonexistent_field", "money")

        data = {"nonexistent_field": "12.00", "money": "10.00"}

        serializer = PropertyModelSerializer(data=data)
        with pytest.raises(ValueError) as ex:
            serializer.is_valid()

        expected_error_message = (
            "nonexistent_field is neither a db field nor a property on the model MoneyFieldModelWithProperty"
        )
        assert ex.value.args[0] == expected_error_message



import django
from django.template import Context, Template, TemplateSyntaxError
from django.utils.translation import override

import pytest

from djmoney.money import Money
from djmoney.templatetags.djmoney import MoneyLocalizeNode


def render(template, context):
    return Template(template).render(Context(context))


class TestMoneyLocalizeNode:
    def test_repr(self):
        assert repr(MoneyLocalizeNode(Money(5, "EUR"))) == "<MoneyLocalizeNode 5 EUR>"

    def test_invalid_instance(self):
        with pytest.raises(Exception) as exc:
            MoneyLocalizeNode(Money(5, "EUR"), amount=15)
        assert str(exc.value) == 'You can define either "money" or the "amount" and "currency".'


@pytest.mark.parametrize(
    "template, context, error_text",
    (
        (
            '{% load djmoney %}{% money_localize "2.5" "PLN" as NEW_M and blabla %}{{NEW_M}}',
            {},
            "Wrong number of input data to the tag.",
        ),
        (
            "{% load djmoney %}{% money_localize money %}{{NEW_M}}",
            {"money": "Something else"},
            'The variable "money" must be an instance of Money.',
        ),
        (
            "{% load djmoney %}{% money_localize amount currency %}",
            {"amount": None, "currency": "PLN"},
            "You must define both variables: amount and currency.",
        ),
    ),
)
def test_invalid_input(template, context, error_text):
    with pytest.raises(TemplateSyntaxError) as exc:
        render(template, context)
    assert str(exc.value) == error_text


def assert_template(string, result, context=None):
    context = context or {}
    with override("pl"):
        assert render(string, context) == result


@pytest.mark.parametrize(
    "string, result, context",
    (
        ('{% load djmoney %}{% money_localize "2.5" "PLN" as NEW_M %}{{NEW_M}}', "2,50\xa0zł", {}),
        ('{% load djmoney %}{% money_localize "2.5" "PLN" %}', "2,50\xa0zł", {}),
        ("{% load djmoney %}{% money_localize amount currency %}", "2,60\xa0zł", {"amount": 2.6, "currency": "PLN"}),
        ("{% load djmoney %}{% money_localize money as NEW_M %}{{NEW_M}}", "2,30\xa0zł", {"money": Money(2.3, "PLN")}),
        (
            "{% load djmoney %}{% money_localize money off as NEW_M %}{{NEW_M}}",
            "2,30\xa0zł",
            {"money": Money(2.3, "PLN")},
        ),
        (
            "{% load djmoney %}{% money_localize money off as NEW_M %}{{NEW_M}}",
            "0,00\xa0zł",
            {"money": Money(0, "PLN")},
        ),
        (
            # with a tag template "money_localize"
            "{% load djmoney %}{% money_localize money %}",
            "2,30\xa0zł",
            {"money": Money(2.3, "PLN")},
        ),
        (
            # without a tag template "money_localize"
            "{{ money }}",
            "2,30\xa0zł",
            {"money": Money(2.3, "PLN")},
        ),
        ("{% load djmoney %}{% money_localize money off %}", "2,30\xa0zł", {"money": Money(2.3, "PLN")}),
        ("{% load djmoney %}{% money_localize money on %}", "2,30\xa0zł", {"money": Money(2.3, "PLN")}),
        (
            # in django 2.0 we fail inside the for loop
            '{% load djmoney %}{% for i in "xxx" %}{% money_localize money %} {% endfor %}',
            "2,30\xa0zł 2,30\xa0zł 2,30\xa0zł ",
            {"money": Money(2.3, "PLN"), "test": "test"},
        ),
    ),
)
def test_tag(string, result, context):
    assert_template(string, result, context)


@pytest.mark.parametrize(
    "string, result, context",
    (
        (
            # money_localize has a default setting USE_L10N = True
            "{% load djmoney %}{% money_localize money %}",
            "2,30\xa0zł",
            {"money": Money(2.3, "PLN")},
        ),
        (
            # without a tag template "money_localize"
            "{{ money }}",
            "2,30\xa0zł",
            {"money": Money(2.3, "PLN")},
        ),
        ("{% load djmoney %}{% money_localize money on %}", "2,30\xa0zł", {"money": Money(2.3, "PLN")}),
    ),
)
def test_l10n_off(settings, string, result, context):
    # This test is only nice to run in older version of Django that do not either
    # complain that the setting is deprecated or as of Django 5.0 entirely ignores
    # this setting
    if django.VERSION < (4, 0):
        settings.USE_L10N = False
        assert_template(string, result, context)


def test_forced_l10n():
    mp = Money(2.3, "PLN")
    mp.use_l10n = True
    assert_template("{{ money }}", "2,30\xa0zł", {"money": mp})



import os
from textwrap import dedent

from django.db import migrations

import pytest

import tests
from djmoney.models.fields import CurrencyField, MoneyField

from .helpers import get_operations


class TestMigrationFramework:
    installed_apps = ["djmoney", "money_app"]
    migration_output = ["*Applying money_app.0001_test... OK*", "*Applying money_app.0002_test... OK*"]

    @pytest.fixture(autouse=True)
    def setup(self, testdir):
        """
        Creates application module, helpers and settings file with basic config.
        """
        self.testdir = testdir
        self.project_root = testdir.mkpydir("money_app")
        testdir.makepyfile(
            app_settings="""
            DATABASES = {
                'default': {
                    'ENGINE': 'django.db.backends.sqlite3',
                    'NAME': 'test.db',
                }
            }
            INSTALLED_APPS = %s
            SECRET_KEY = 'foobar'
            """
            % str(self.installed_apps)
        )
        self.project_root.join("migrations/__init__.py").ensure()
        testdir.syspathinsert()

    def make_models(self, content):
        """
        Creates models.py file.
        """
        fd = self.project_root.join("models.py")
        fd.write(dedent(content))

    def make_migration(self, callable_default='Money("0.00 USD")', **fields):
        """
        Creates a model with provided fields and creates a migration for it.
        """
        if fields:
            fields_definition = ";".join("=".join([field, definition]) for field, definition in fields.items())
        else:
            fields_definition = "pass"
        self.make_models(
            """
            from django.db import models

            from djmoney.models.fields import MoneyField, Money

            # A callable default
            def default1():
                return {callable_default}

            class Model(models.Model):
                {fields}
            """.format(
                fields=fields_definition, callable_default=callable_default
            )
        )
        tests_path = os.path.dirname(os.path.dirname(tests.__file__))
        return self.run(
            f"""
            import sys
            sys.path.append('{tests_path}')
            from tests.migrations.helpers import makemigrations
            makemigrations()
            """
        )

    def make_default_migration(self, field="MoneyField(max_digits=10, decimal_places=2, null=True)"):
        return self.make_migration(field=field)

    def run(self, content):
        return self.testdir.runpython_c(
            dedent(
                """
                import os
                os.environ['DJANGO_SETTINGS_MODULE'] = 'app_settings'
                from django import setup

                setup()
                """
            )
            + dedent(content)
        )

    def create_instance(self):
        self.run(
            """
            from money_app.models import Model
            from djmoney.money import Money

            Model.objects.create(field=Money(10, 'USD'))
            """
        )

    def migrate(self):
        tests_path = os.path.dirname(os.path.dirname(tests.__file__))
        return self.run(
            dedent(
                f"""
                import sys
                sys.path.append('{tests_path}')
                from tests.migrations.helpers import migrate
                migrate()
                """
            )
        )

    def assert_migrate(self, output=None):
        """
        Runs migrations and checks if 2 migrations were applied.
        """
        if output is None:
            output = self.migration_output
        migration = self.migrate()
        migration.stdout.fnmatch_lines(output)

    def test_create_initial(self):
        migration = self.make_default_migration()
        print(migration.stdout)
        # Django 5 output replaced "-" with "+"
        # Migrations for 'money_app':
        #  money_app/migrations/0001_test.py
        #    + Create model Model
        migration.stdout.fnmatch_lines(["*Migrations for 'money_app':*", "*0001_test.py*", "*Create model Model*"])
        operations = get_operations("0001")
        assert len(operations) == 1

        assert isinstance(operations[0], migrations.CreateModel)

        fields = sorted(operations[0].fields)
        assert len(fields) == 3
        assert fields[0][0] == "field"
        assert isinstance(fields[0][1], MoneyField)
        assert fields[1][0] == "field_currency"
        assert isinstance(fields[1][1], CurrencyField)
        migration = self.migrate()
        migration.stdout.fnmatch_lines(["*Applying money_app.0001_test... OK*"])

    def test_add_field(self):
        self.make_migration()
        migration = self.make_default_migration()
        migration.stdout.fnmatch_lines(
            [
                "*Migrations for 'money_app':*",
                "*0002_test.py*",
                "*Add field field to model*",
                "*Add field field_currency to model*",
            ]
        )

        operations = get_operations("0002")
        assert len(operations) == 2
        assert isinstance(operations[0], migrations.AddField)
        assert isinstance(operations[0].field, MoneyField)
        assert isinstance(operations[1], migrations.AddField)
        assert isinstance(operations[1].field, CurrencyField)
        self.assert_migrate()

    def test_alter_field(self):
        self.make_default_migration()
        migration = self.make_migration(field="MoneyField(max_digits=15, decimal_places=2, null=True)")
        migration.stdout.fnmatch_lines(
            ["*Migrations for 'money_app':*", "*0002_test.py*", "*Alter field field on model*"]
        )

        operations = get_operations("0002")
        assert len(operations) == 1
        assert isinstance(operations[0], migrations.AlterField)
        assert isinstance(operations[0].field, MoneyField)
        assert operations[0].field.max_digits == 15
        self.assert_migrate()

    def test_alter_field_callable(self):
        """Test that changing the return value of a callable does not produce new migrations."""
        self.make_default_migration()
        # Change the field to have a callable default
        migration = self.make_migration(
            field="MoneyField(max_digits=15, decimal_places=2, null=True, default=default1)"
        )
        migration.stdout.fnmatch_lines(
            ["*Migrations for 'money_app':*", "*0002_test.py*", "*Alter field field on model*"]
        )

        # Change the return value of the callable
        migration = self.make_migration(
            callable_default='"USD"', field="MoneyField(max_digits=15, decimal_places=2, null=True, default=default1)"
        )
        migration.stdout.fnmatch_lines(["No changes detected in app 'money_app'"])

    def test_alter_field_currency_callable(self):
        """Test that changing the return value of a callable does not produce new migrations."""
        self.make_default_migration()
        # Change the field to have a callable default
        migration = self.make_migration(
            field="MoneyField(max_digits=15, decimal_places=2, null=True, default_currency=default1)"
        )
        migration.stdout.fnmatch_lines(
            ["*Migrations for 'money_app':*", "*0002_test.py*", "*Alter field field on model*"]
        )

        # Change the return value of the callable
        migration = self.make_migration(
            callable_default='"USD"',
            field="MoneyField(max_digits=15, decimal_places=2, null=True, default_currency=default1)",
        )
        migration.stdout.fnmatch_lines(["No changes detected in app 'money_app'"])

    def test_alter_field_currency_choices_callable(self):
        """Test that changing the return value of a callable does not produce new migrations."""
        self.make_default_migration()
        # Change the field to have a callable default
        migration = self.make_migration(
            callable_default='[("USD", "USD")]',
            field="MoneyField(max_digits=15, decimal_places=2, null=True, currency_choices=default1)",
        )
        migration.stdout.fnmatch_lines(
            ["*Migrations for 'money_app':*", "*0002_test.py*", "*Alter field field on model*"]
        )

        # Change the return value of the callable
        migration = self.make_migration(
            callable_default='[("CNY", "CNY")]',
            field="MoneyField(max_digits=15, decimal_places=2, null=True, currency_choices=default1)",
        )
        migration.stdout.fnmatch_lines(["No changes detected in app 'money_app'"])

    def test_remove_field(self):
        self.make_default_migration()
        migration = self.make_migration()
        migration.stdout.fnmatch_lines(
            [
                "*Migrations for 'money_app':*",
                "*0002_test.py*",
                "*Remove field field from model*",
                "*Remove field field_currency from model*",
            ]
        )

        operations = get_operations("0002")
        assert len(operations) == 2
        assert isinstance(operations[0], migrations.RemoveField)
        assert operations[0].name == "field"
        assert isinstance(operations[1], migrations.RemoveField)
        assert operations[1].name == "field_currency"
        self.assert_migrate()

    def test_rename_field(self):
        self.make_default_migration()
        self.assert_migrate(["*Applying money_app.0001_test... OK*"])
        self.create_instance()
        migration = self.make_migration(new_field="MoneyField(max_digits=10, decimal_places=2, null=True)")
        migration.stdout.fnmatch_lines(
            [
                "*Migrations for 'money_app':*",
                "*0002_test.py*",
                "*Rename field field on model to new_field*",
                "*Rename field field_currency on model to new_field_currency*",
            ]
        )
        self.assert_migrate(["*Applying money_app.0002_test... OK*"])
        result = self.run(
            """
            from money_app.models import Model

            instance = Model.objects.get()
            print(instance.new_field)
            """
        )
        result.stdout.fnmatch_lines(["$10.00"])

    def test_migrate_to_moneyfield(self):
        self.make_default_migration(field="models.DecimalField(max_digits=10, decimal_places=2, null=True)")
        migration = self.make_migration(field="MoneyField(max_digits=10, decimal_places=2, null=True)")
        migration.stdout.fnmatch_lines(
            ["*Migrations for 'money_app':*", "*0002_test.py*", "*Add field field_currency to model*"]
        )
        self.assert_migrate()

    def test_migrate_from_moneyfield(self):
        self.make_default_migration()
        migration = self.make_migration(field="models.DecimalField(max_digits=10, decimal_places=2, null=True)")
        migration.stdout.fnmatch_lines(
            ["*Migrations for 'money_app':*", "*0002_test.py*", "*Remove field field_currency from model*"]
        )
        self.assert_migrate()



# flake8: noqa
"""
This module isolates code that has to do with compatibility issues between
different supported versions of Python or other dependencies.

This is quite important to keep the codebase clean.

Please do not catch ImportError exceptions other places than here :)
"""


def setup_managers(sender):
    from .models.managers import money_manager

    default_manager_name = sender._meta.default_manager_name or "objects"
    for manager in filter(lambda m: m.name == default_manager_name, sender._meta.local_managers):
        money_manager(manager)



from django.utils.module_loading import import_string

from djmoney.contrib.exchange.models import Rate

from ..base import BaseExchangeCommand


class Command(BaseExchangeCommand):
    help = "Clears exchange rates."

    def add_arguments(self, parser):
        super().add_arguments(parser)
        parser.add_argument(
            "--all",
            action="store_true",
            dest="all",
            help="Clear rates for all backends.",
            required=False,
            default=False,
        )

    def handle(self, *args, **options):
        if options["all"]:
            Rate.objects.all().delete()
            message = "Successfully cleared all rates"
        else:
            backend = import_string(options["backend"])
            Rate.objects.filter(backend=backend.name).delete()
            message = "Successfully cleared rates for %s" % backend.name
        self.success(message)



from django.db.models import F, Q

import pytest

from djmoney.models.managers import _expand_money_args, _expand_money_kwargs
from djmoney.money import Money
from djmoney.utils import get_amount
from moneyed import Money as OldMoney

from .testapp.models import ModelWithNonMoneyField


def assert_leafs(children):
    assert ("money", Money(0, "USD")) in children
    assert ("money_currency", "USD") in children


class TestExpandMoneyArgs:
    def test_no_args(self):
        assert _expand_money_args(ModelWithNonMoneyField(), []) == []

    def test_non_q_args(self):
        assert _expand_money_args(ModelWithNonMoneyField(), ["money"]) == ["money"]

    def test_exact(self):
        """
        Test
          (AND: ('money', 0 USD))
        results in;
          (AND: (AND: ('money', 0 USD), ('money_currency', u'USD')))
        """
        actual = _expand_money_args(ModelWithNonMoneyField(), [Q(money=Money(0, "USD"))])

        assert len(actual) == 1
        arg = actual[0]

        assert isinstance(arg, Q)
        assert arg.connector == Q.AND
        assert len(arg.children) == 1
        assert isinstance(arg.children[0], Q)
        assert arg.children[0].connector == Q.AND
        assert_leafs(arg.children[0].children)

    def test_and(self):
        """
        Test
          (AND: ('desc', 'foo'), ('money', 0 USD))
        results in;
          (AND: ('desc', 'foo'), (AND: ('money', 0 USD), ('money_currency', u'USD')))
        """
        actual = _expand_money_args(ModelWithNonMoneyField(), [Q(money=Money(0, "USD"), desc="foo")])

        assert len(actual) == 1
        arg = actual[0]

        assert isinstance(arg, Q)
        assert arg.connector == Q.AND
        assert len(arg.children) == 2

        # Can't guarantee the ordering of children, thus;
        for child in arg.children:
            if isinstance(child, tuple):
                assert ("desc", "foo") == child
            elif isinstance(child, Q):
                assert child.connector == Q.AND
                assert len(child.children) == 2
                assert_leafs(child.children)
            else:
                pytest.fail("There should only be two elements, a tuple and a Q - not a %s" % child)

    def test_and_with_or(self):
        """
        Test
          (OR: (AND: ('desc', 'foo'), ('money', 0 USD)), ('desc', 'bar'))
        results in:
          (OR: (AND: ('desc', 'foo'), (AND: ('money', 0 USD), ('money_currency', u'USD'))), ('desc', 'bar'))
        """
        actual = _expand_money_args(ModelWithNonMoneyField(), [Q(money=Money(0, "USD"), desc="foo") | Q(desc="bar")])

        assert len(actual) == 1
        arg = actual[0]

        assert isinstance(arg, Q)
        assert arg.connector == Q.OR
        assert len(arg.children) == 2

        # Can't guarantee the ordering of children, thus;
        for child in arg.children:
            if isinstance(child, tuple):
                assert ("desc", "bar") == child
            elif isinstance(child, Q):
                assert len(child.children) == 2
                for subchild in child.children:
                    if isinstance(subchild, tuple):
                        assert ("desc", "foo") == subchild
                    elif isinstance(subchild, Q):
                        assert_leafs(subchild.children)
            else:
                pytest.fail("There should only be two elements, a tuple and a Q - not a %s" % child)

    def test_and_with_two_or(self):
        """
        Test
          (OR:
            (OR:
              (AND:
                ('desc', 'foo'),
                ('money', 0 USD)
              ),
              ('desc', 'eggs')
            ),
            ('desc', 'bar')
          )
        results in;
          (OR:
            (OR:
              (AND:
                ('desc', 'foo'),
                (AND:
                  ('money', 0),
                  ('money_currency', 'USD')
                )
              ),
              ('desc', 'eggs')
            ),
            ('desc', 'bar')
          )
        """
        actual = _expand_money_args(
            ModelWithNonMoneyField(), [Q(Q(money=Money(0, "USD"), desc="foo") | Q(desc="eggs")) | Q(desc="bar")]
        )
        arg = actual[0]

        assert len(actual) == 1
        assert isinstance(arg, Q)
        assert arg.connector == Q.OR
        assert len(arg.children) == 2

        # Can't guarantee the ordering of children, thus;
        for child in arg.children:
            if isinstance(child, tuple):
                assert ("desc", "bar") == child
            elif isinstance(child, Q):
                assert len(child.children) == 2
                for subchild in child.children:
                    if isinstance(subchild, tuple):
                        assert ("desc", "eggs") == subchild
                    elif isinstance(subchild, Q):
                        for subsubchild in subchild.children:
                            if isinstance(subsubchild, tuple):
                                assert ("desc", "foo") == subsubchild
                            elif isinstance(subsubchild, Q):
                                assert_leafs(subsubchild.children)
                            else:
                                pytest.fail(
                                    "There should only be two subsubchild elements, a tuple and a Q - not a %s"
                                    % subsubchild
                                )
                    else:
                        pytest.fail(
                            "There should only be two subchild elements, a tuple and a Q - not a %s" % subsubchild
                        )
            else:
                pytest.fail("There should only be two child elements, a tuple and a Q - not a %s" % child)


class TestKwargsExpand:
    @pytest.mark.parametrize(
        "value, expected",
        (
            (
                ({"money": 100, "desc": "test"}, {"money": 100, "desc": "test"}),
                ({"money": Money(100, "USD")}, {"money": 100, "money_currency": "USD"}),
                ({"money": OldMoney(100, "USD")}, {"money": 100, "money_currency": "USD"}),
                ({"money": Money(100, "USD"), "desc": "test"}, {"money": 100, "money_currency": "USD", "desc": "test"}),
            )
        ),
    )
    def test_simple(self, value, expected):
        assert _expand_money_kwargs(ModelWithNonMoneyField, kwargs=value)[1] == expected

    @pytest.mark.parametrize(
        "value, expected", (({"money": F("money") * 2}, 2), ({"money": F("money") + Money(100, "USD")}, 100))
    )
    def test_complex_f_query(self, value, expected):
        _, kwargs = _expand_money_kwargs(ModelWithNonMoneyField, kwargs=value)
        assert isinstance(kwargs["money_currency"], F)
        assert kwargs["money_currency"].name == "money_currency"
        assert get_amount(kwargs["money"].rhs) == expected

    def test_simple_f_query(self):
        _, kwargs = _expand_money_kwargs(ModelWithNonMoneyField, kwargs={"money": F("money")})
        assert isinstance(kwargs["money_currency"], F)
        assert kwargs["money_currency"].name == "money_currency"
        assert kwargs["money"].name == "money"



import django
import django.contrib.admin.utils as admin_utils
from django.urls import reverse

import pytest

from djmoney.money import Money

from .testapp.models import ModelWithParentAndCallableFields, ModelWithVanillaMoneyField, ParentModel


MONEY_FIELD = ModelWithVanillaMoneyField._meta.get_field("money")
INTEGER_FIELD = ModelWithVanillaMoneyField._meta.get_field("integer")


@pytest.mark.parametrize(
    "value, expected",
    (
        (Money(10, "RUB"), "10,00\xa0RUB"),  # Issue 232
        (Money(1000, "SAR"), "1\xa0000,00\xa0SAR"),  # Issue 196
        (Money(1000, "PLN"), "1\xa0000,00\xa0PLN"),  # Issue 102
        (Money("3.33", "EUR"), "3,33\xa0€"),  # Issue 90
    ),
)
def test_display_for_field(settings, value, expected):
    # This now defaults to True and raises RemovedInDjango50Warning
    if django.VERSION < (4, 0):
        settings.USE_L10N = True
    # This locale has no definitions in py-moneyed, so it will work for localized money representation.
    settings.LANGUAGE_CODE = "cs"
    assert admin_utils.display_for_field(value, MONEY_FIELD, "") == expected


def test_default_display():
    assert admin_utils.display_for_field(10, INTEGER_FIELD, "") == "10"


def test_admin_with_formset(admin_user, admin_client):
    """Test to assure that formsets with MoneyFields behave correctly"""

    parent = ParentModel.objects.create()
    obj = ModelWithParentAndCallableFields.objects.create(parent=parent)

    url = reverse(
        "admin:testapp_parentmodel_change",
        kwargs={
            "object_id": obj.pk,
        },
    )
    response = admin_client.get(url)
    assert response.status_code == 200

    response = admin_client.post(
        url,
        {
            "_save": "Save",
            "modelwithparentandcallablefields_set-TOTAL_FORMS": 2,
            "modelwithparentandcallablefields_set-INITIAL_FORMS": 1,
            "modelwithparentandcallablefields_set-MIN_NUM_FORMS": 0,
            "modelwithparentandcallablefields_set-MAX_NUM_FORMS": 2,
            # The first object exists. We sent back updated values.
            "modelwithparentandcallablefields_set-0-modelwithcallabledefaultanddefaultcurrency_ptr": obj.pk,
            "modelwithparentandcallablefields_set-0-parent": 1,
            "modelwithparentandcallablefields_set-0-money_0": 123,
            "modelwithparentandcallablefields_set-0-money_1": "CHF",
            # This is a very sensitive field value, so we are also checking that it's actually part of the generated
            # form data.
            "initial-modelwithparentandcallablefields_set-0-money": "0.00 EUR",
            # We want this to be all default values and check that it is NOT created
            "modelwithparentandcallablefields_set-1-modelwithcallabledefaultanddefaultcurrency_ptr": "",
            "modelwithparentandcallablefields_set-1-parent": 1,
            "modelwithparentandcallablefields_set-1-money_0": "0.00",
            "modelwithparentandcallablefields_set-1-money_1": "EUR",
            "initial-modelwithparentandcallablefields_set-1-money": "0.00 EUR",
        },
    )

    print(response.content)

    assert response.status_code == 302



import warnings

import django

import moneyed


DATABASES = {"default": {"ENGINE": "django.db.backends.sqlite3", "NAME": ":memory:"}}

TEMPLATES = [
    {
        "BACKEND": "django.template.backends.django.DjangoTemplates",
        "DIRS": [],
        "APP_DIRS": True,
        "OPTIONS": {"context_processors": ["django.contrib.auth.context_processors.auth"]},
    }
]

MIDDLEWARE = [
    "django.middleware.security.SecurityMiddleware",
    "django.contrib.sessions.middleware.SessionMiddleware",
    "django.middleware.locale.LocaleMiddleware",
    "django.middleware.common.CommonMiddleware",
    "django.middleware.csrf.CsrfViewMiddleware",
    "django.contrib.auth.middleware.AuthenticationMiddleware",
    "django.contrib.messages.middleware.MessageMiddleware",
    "django.middleware.clickjacking.XFrameOptionsMiddleware",
]

warnings.simplefilter("ignore", Warning)

INSTALLED_APPS = [
    "django.contrib.admin",
    "django.contrib.auth",
    "django.contrib.contenttypes",
    "django.contrib.sessions",
    "django.contrib.sites",
    "djmoney",
    "djmoney.contrib.exchange",
    "tests.testapp",
]

SITE_ID = 1

SECRET_KEY = "foobar"

ROOT_URLCONF = "tests.urls"

# This now defaults to True and raises RemovedInDjango50Warning
if django.VERSION < (4, 0):
    USE_L10N = True


moneyed.add_currency(code="USDT", numeric="000", name="Tether")

OPEN_EXCHANGE_RATES_APP_ID = "test"
FIXER_ACCESS_KEY = "test"
UNIRATE_ACCESS_KEY = "test"

DEFAULT_CURRENCY = "USD"

# Additional setting for callables with different defaults
TEST_DEFAULT_CURRENCY = "EUR"



"""
This module contains various helpers for migrations testing.
"""

import os


MIGRATION_NAME = "test"


def makemigrations():
    from django.core.management import call_command
    from django.db.migrations import questioner

    # We should answer yes for all migrations questioner questions
    questioner.input = lambda prompt=None: "y"

    os.system("find . -name \\*.pyc -delete")
    call_command("makemigrations", "money_app", name=MIGRATION_NAME)


def get_migration(name):
    return __import__(f"money_app.migrations.{name}_{MIGRATION_NAME}", fromlist=["Migration"]).Migration


def get_operations(migration_name):
    return get_migration(migration_name).operations


def migrate():
    from django.core.management import call_command

    call_command("migrate", "money_app")



import django.contrib.admin.helpers as admin_helpers
import django.contrib.admin.templatetags.admin_list as admin_list
import django.contrib.admin.utils as admin_utils

from .models.fields import MoneyField


MODULES_TO_PATCH = [admin_utils, admin_helpers, admin_list]


def setup_admin_integration():
    original_display_for_field = admin_utils.display_for_field

    def display_for_field(value, field, *args, **kwargs):
        if isinstance(field, MoneyField):
            return str(value)
        return original_display_for_field(value, field, *args, **kwargs)

    for mod in MODULES_TO_PATCH:
        setattr(mod, "display_for_field", display_for_field)



from __future__ import annotations

from decimal import Decimal
from typing import Any, Optional, Union, overload, type_check_only

from django.db.models.expressions import CombinedExpression, F
from django.utils.safestring import SafeText

from typing_extensions import Literal, TypedDict

from moneyed import Currency, Money as _DefaultMoney

__all__ = ["Money", "Currency"]
DefaultMoney = _DefaultMoney

@type_check_only
class _FormatOptions(TypedDict):
    "Format options parameter."

    format: Optional[str]
    locale: str
    currency_digits: bool
    format_type: Literal["standard", "standard:short", "name"]
    decimal_quantization: bool

class Money(DefaultMoney):
    use_l10n: Any
    decimal_places: Any
    format_options: Any

    def __init__(self, *args, format_options: Optional[_FormatOptions] = ..., **kwargs) -> None: ...
    @overload  # type: ignore[override]
    def __add__(self, other: Money) -> Money: ...
    @overload
    def __add__(self, other: F) -> CombinedExpression: ...
    @overload  # type: ignore[override]
    def __sub__(self, other: Money) -> Money: ...
    @overload
    def __sub__(self, other: F) -> CombinedExpression: ...
    @overload  # type: ignore[override]
    def __mul__(self, other: _NumericalType) -> Money: ...
    @overload
    def __mul__(self, other: F) -> CombinedExpression: ...
    @overload  # type: ignore[override]
    def __truediv__(self, other: _NumericalType) -> Money: ...
    @overload
    def __truediv__(self, other: F) -> CombinedExpression: ...
    def __rtruediv__(self, other) -> None: ...  # type: ignore[override]
    @property
    def is_localized(self) -> bool: ...
    def __html__(self) -> SafeText: ...
    def __round__(self, n: Any | None = ...) -> Money: ...
    def round(self, ndigits: int = ...) -> Money: ...  # type: ignore[override]
    def __pos__(self) -> Money: ...
    def __neg__(self) -> Money: ...
    def __abs__(self) -> Money: ...
    def __rmod__(self, other) -> Money: ...
    def __radd__(self, other) -> Money: ...  # type: ignore[override]
    def __rmul__(self, other) -> Money: ...  # type: ignore[override]

def get_current_locale() -> str: ...

_NumericalType = Union[Decimal, float, int]



from django.core.management.base import BaseCommand

from djmoney import settings


class BaseExchangeCommand(BaseCommand):
    """
    Basic command for exchange rates manipulation.
    Provides ``backend`` argument.
    """

    def add_arguments(self, parser):
        parser.add_argument(
            "-b",
            "--backend",
            action="store",
            dest="backend",
            help="Importable string for custom exchange rates backend.",
            required=False,
            default=settings.EXCHANGE_BACKEND,
        )

    def success(self, message):
        self.stdout.write(self.style.SUCCESS(message))



from decimal import Decimal

from django.core.exceptions import ValidationError
from django.core.validators import BaseValidator
from django.utils.translation import gettext_lazy as _

from djmoney.money import Money


class BaseMoneyValidator(BaseValidator):
    def get_limit_value(self, cleaned):
        limit_value = self.limit_value() if callable(self.limit_value) else self.limit_value
        if isinstance(limit_value, Money):
            if cleaned.currency.code != limit_value.currency.code:
                return
            return limit_value
        elif isinstance(limit_value, (int, Decimal)):
            return limit_value
        try:
            return Money(limit_value[cleaned.currency.code], cleaned.currency.code)
        except KeyError:
            # There are no validation for this currency
            pass

    def __call__(self, value):
        cleaned = self.clean(value)
        limit_value = self.get_limit_value(cleaned)
        if limit_value is None:
            return
        if isinstance(limit_value, (int, Decimal)):
            cleaned = cleaned.amount
        params = {"limit_value": limit_value, "show_value": cleaned, "value": value}
        if self.compare(cleaned, limit_value):
            raise ValidationError(self.message, code=self.code, params=params)


# The validators below are not inherited from Django's validators because of
# subclass check in Django REST Framework, that removes all validators
# from the serializer field except the first one in the `validators` list of each validator type (min / max)
# Removing inheritance is the simplest option to make model validators work in REST Framework


class MinMoneyValidator(BaseMoneyValidator):
    message = _("Ensure this value is greater than or equal to %(limit_value)s.")
    code = "min_value"

    def compare(self, a, b):
        return a < b


class MaxMoneyValidator(BaseMoneyValidator):
    message = _("Ensure this value is less than or equal to %(limit_value)s.")
    code = "max_value"

    def compare(self, a, b):
        return a > b



"""
Created on May 7, 2011

@author: jake
"""

import warnings
from decimal import Decimal

import pytest

from djmoney import settings
from djmoney.models.fields import MoneyField
from djmoney.money import Money

from .testapp.forms import (
    DefaultMoneyModelForm,
    DefaultPrecisionModelForm,
    DisabledFieldForm,
    MoneyForm,
    MoneyFormMultipleCurrencies,
    MoneyModelForm,
    MoneyModelFormWithCallableCurrencyChoices,
    MoneyModelFormWithCallableDefault,
    MoneyModelFormWithCallableDefaultAndCallableDefaultCurrency,
    MoneyModelFormWithCallableDefaultCurrency,
    MoneyModelFormWithValidation,
    NullableModelForm,
    OptionalMoneyForm,
    PositiveValidatedMoneyModelForm,
    PreciseForm,
    PreciseModelForm,
    ValidatedMoneyModelForm,
)
from .testapp.models import (
    ModelWithCallableCurrencyChoices,
    ModelWithCallableDefault,
    ModelWithCallableDefaultAndDefaultCurrency,
    ModelWithCallableDefaultCurrency,
    ModelWithVanillaMoneyField,
    NullMoneyFieldModel,
)


pytestmark = pytest.mark.django_db


def test_save():
    money = Money(Decimal("10"), "SEK")
    form = MoneyModelForm({"money_0": money.amount, "money_1": money.currency})

    assert form.is_valid()
    instance = form.save()

    retrieved = ModelWithVanillaMoneyField.objects.get(pk=instance.pk)
    assert money == retrieved.money


def test_save_with_callable_default():
    """Test if the callable default triggers"""
    expected_money = Money(Decimal("0"), "EUR")
    form = MoneyModelFormWithCallableDefault({})

    assert form.is_valid()
    instance = form.save()

    retrieved = ModelWithCallableDefault.objects.get(pk=instance.pk)
    assert expected_money == retrieved.money


def test_save_with_callable_default_currency():
    """Test if the callable default triggers"""
    expected_money = Money(Decimal("10"), "EUR")
    form = MoneyModelFormWithCallableDefaultCurrency({"money_0": "10", "money_1": "EUR"})

    assert form.fields["money"].initial[0] is None
    assert form.fields["money"].initial[1] == "EUR"
    assert """<option value="EUR" selected>""" in form.as_p()
    assert form.is_valid()

    instance = form.save()

    retrieved = ModelWithCallableDefaultCurrency.objects.get(pk=instance.pk)
    assert expected_money == retrieved.money


def test_save_with_callable_default_and_currency():
    """Test if the callable defaults trigger when both are defined"""
    expected_money = Money(Decimal("0"), "EUR")
    form = MoneyModelFormWithCallableDefaultAndCallableDefaultCurrency({})

    assert form.is_valid()
    instance = form.save()

    retrieved = ModelWithCallableDefaultAndDefaultCurrency.objects.get(pk=instance.pk)
    assert expected_money == retrieved.money


def test_save_with_callable_currency_choices():
    """Test if the callable default triggers"""
    expected_money = Money(Decimal("10"), "EUR")
    form = MoneyModelFormWithCallableCurrencyChoices({"money_0": "10", "money_1": "EUR"})

    assert form.fields["money"].initial[0] is None
    assert form.fields["money"].initial[1] == "EUR"
    assert list(form.fields["money"].widget.widgets[1].choices) == [("DKK", "DKK"), ("EUR", "EUR"), ("USD", "USD")]
    assert """<option value="EUR" selected>""" in form.as_p()
    assert form.is_valid()

    instance = form.save()

    retrieved = ModelWithCallableCurrencyChoices.objects.get(pk=instance.pk)
    assert expected_money == retrieved.money


def test_validate():
    money = Money(Decimal("10"), "SEK")
    form = MoneyForm({"money_0": money.amount, "money_1": money.currency})

    assert form.is_valid()

    result = form.cleaned_data["money"]
    assert result == money


@pytest.mark.parametrize(
    "data",
    (
        {"money_0": "xyz*|\\", "money_1": "SEK"},
        {"money_0": 10000, "money_1": "SEK"},
        {"money_0": 1, "money_1": "SEK"},
        {"money_0": 10, "money_1": "EUR"},
    ),
)
def test_form_is_invalid(data):
    assert not MoneyForm(data).is_valid()


@pytest.mark.parametrize(
    "data, result", (({"money_0": "", "money_1": "SEK"}, []), ({"money_0": "1.23", "money_1": "SEK"}, ["money"]))
)
def test_changed_data(data, result):
    assert MoneyForm(data).changed_data == result


def test_change_currency_not_amount():
    """
    If the amount is the same, but the currency changes, then we
    should consider this to be a change.
    """
    form = MoneyFormMultipleCurrencies(
        {"money_0": Decimal(10), "money_1": "EUR"}, initial={"money": Money(Decimal(10), "SEK")}
    )
    assert form.changed_data == ["money"]


@pytest.mark.parametrize(
    "data, result",
    (
        ({"money_1": "SEK"}, True),
        ({"money_0": "", "money_1": "SEK"}, True),
        ({"money_0": "xyz*|\\", "money_1": "SEK"}, False),
    ),
)
def test_optional_money_form(data, result):
    """
    The currency widget means that 'money_1' will always be filled
    in, but 'money_0' could be absent/empty.
    """
    assert OptionalMoneyForm(data).is_valid() is result


def test_default_currency():
    """
    If field is nullable, then field's default_currency value should be selected by default.
    """
    instance = NullMoneyFieldModel.objects.create()
    form = NullableModelForm(instance=instance)
    assert '<option value="USD" selected>US Dollar</option>' in form.as_p()


def test_fields_default_amount_becomes_forms_initial():
    """
    Formfield should take field's default amount
    and put it in form field's initial value
    """
    form = DefaultMoneyModelForm()
    assert form.fields["money"].initial == [123, "PLN"]


def test_no_deprecation_warning():
    """
    The library's code shouldn't generate any warnings itself. See #262.
    """
    with warnings.catch_warnings():
        warnings.simplefilter("error")
        MoneyField(max_digits=10, decimal_places=2, currency_choices=(("USD", "USD"),)).formfield()


class TestValidation:
    @pytest.mark.parametrize(
        "value, error",
        (
            (Money(50, "EUR"), "Ensure this value is greater than or equal to €100.00."),
            (Money(1500, "EUR"), "Ensure this value is less than or equal to €1,000.00."),
            (Money(40, "USD"), "Ensure this value is greater than or equal to $50.00."),
            (Money(600, "USD"), "Ensure this value is less than or equal to $500.00."),
            (Money(400, "NOK"), "Ensure this value is greater than or equal to NOK500.00."),
            (Money(950, "NOK"), "Ensure this value is less than or equal to NOK900.00."),
            (Money(5, "SEK"), "Ensure this value is greater than or equal to 10."),
            (Money(1600, "SEK"), "Ensure this value is less than or equal to 1500."),
        ),
    )
    def test_invalid(self, value, error):
        form = ValidatedMoneyModelForm(data={"money_0": value.amount, "money_1": value.currency})
        assert not form.is_valid()
        assert form.errors == {"money": [error]}

    @pytest.mark.parametrize("value", (Money(150, "EUR"), Money(200, "USD"), Money(50, "SEK"), Money(600, "NOK")))
    def test_valid(self, value):
        assert ValidatedMoneyModelForm(data={"money_0": value.amount, "money_1": value.currency}).is_valid()

    @pytest.mark.parametrize("value", (Money(-0.01, "EUR"), Money(-1, "USD"), Money(-10, "NOK"), Money(-100, "SEK")))
    def test_non_negative_validator(self, value):
        """Fails if Validator(0) silently allows negative values."""
        form = PositiveValidatedMoneyModelForm(data={"money_0": value.amount, "money_1": value.currency})
        assert not form.is_valid()
        assert form.errors == {"money": ["Ensure this value is greater than or equal to 0."]}

    @pytest.mark.parametrize(
        "value", (Money(0, "PHP"), Money(0.01, "EUR"), Money(1, "USD"), Money(10, "NOK"), Money(100, "SEK"))
    )
    def test_positive_validator(self, value):
        """Fails if MinMoneyValidator(0) blocks positive values.

        MinMoneyValidator(0) should also allow exactly 0.
        """
        form = PositiveValidatedMoneyModelForm(data={"money_0": value.amount, "money_1": value.currency})
        assert form.is_valid()

    def test_default_django_validator(self):
        form = MoneyModelFormWithValidation(data={"balance_0": 0, "balance_1": "GBP"})
        assert not form.is_valid()
        assert form.errors == {"balance": ["Ensure this value is greater than or equal to £100.00."]}


class TestDisabledField:
    def test_validation(self):
        instance = ModelWithVanillaMoneyField.objects.create(money=Money("42.00", "USD"))
        form = DisabledFieldForm(data={}, instance=instance)
        assert not form.errors
        assert form.is_valid()

    def test_has_changed(self):
        form = DisabledFieldForm(data={})
        assert not form.has_changed()


@pytest.mark.parametrize("model_class", (PreciseForm, DefaultPrecisionModelForm))
def test_decimal_places_model_form(model_class):
    """Forms should use DECIMAL_PLACES setting value when none specified."""

    expected = str(10**-settings.DECIMAL_PLACES)
    assert model_class().fields["money"].widget.widgets[0].attrs["step"] == expected


def test_precedence_decimal_places_model_form():
    """Forms should use decimal_places in field value when specified."""

    decimal_places = PreciseModelForm.Meta.model._meta.fields[2].decimal_places
    expected = str(10**-decimal_places)
    assert PreciseModelForm().fields["money"].widget.widgets[0].attrs["step"] == expected



import warnings
from types import MappingProxyType

from django.conf import settings
from django.db.models import F
from django.utils import translation
from django.utils.deconstruct import deconstructible
from django.utils.html import avoid_wrapping, conditional_escape
from django.utils.safestring import mark_safe

import moneyed.l10n
from moneyed import Currency, Money as DefaultMoney

from .settings import DECIMAL_PLACES, MONEY_FORMAT


__all__ = ["Money", "Currency"]


@deconstructible
class Money(DefaultMoney):
    """
    Extends functionality of Money with Django-related features.
    """

    use_l10n = None

    def __init__(self, *args, format_options=None, **kwargs):
        self.decimal_places = kwargs.pop("decimal_places", DECIMAL_PLACES)
        self.format_options = MappingProxyType(format_options) if format_options is not None else None
        super().__init__(*args, **kwargs)

    def _copy_attributes(self, source, target):
        """Copy attributes to the new `Money` instance.

        This class stores extra bits of information about string formatting that the parent class doesn't have.
        The problem is that the parent class creates new instances of `Money` without in some of its methods and
        it does so without knowing about `django-money`-level attributes.
        For this reason, when this class uses some methods of the parent class that have this behavior, the resulting
        instances lose those attribute values.

        When it comes to what number of decimal places to choose, we take the maximum number.
        """
        selection = [
            getattr(candidate, "decimal_places", None)
            for candidate in (self, source)
            if getattr(candidate, "decimal_places", None) is not None
        ]
        if selection:
            target.decimal_places = max(selection)

    def __add__(self, other):
        if isinstance(other, F):
            return other.__radd__(self)
        other = maybe_convert(other, self.currency)
        result = super().__add__(other)
        self._copy_attributes(other, result)
        return result

    def __sub__(self, other):
        if isinstance(other, F):
            return other.__rsub__(self)
        other = maybe_convert(other, self.currency)
        result = super().__sub__(other)
        self._copy_attributes(other, result)
        return result

    def __mul__(self, other):
        if isinstance(other, F):
            return other.__rmul__(self)
        result = super().__mul__(other)
        self._copy_attributes(other, result)
        return result

    def __truediv__(self, other):
        if isinstance(other, F):
            return other.__rtruediv__(self)
        result = super().__truediv__(other)
        if isinstance(result, self.__class__):
            self._copy_attributes(other, result)
        return result

    def __rtruediv__(self, other):
        # Backported from py-moneyed, non released bug-fix
        # https://github.com/py-moneyed/py-moneyed/blob/c518745dd9d7902781409daec1a05699799474dd/moneyed/classes.py#L217-L218
        raise TypeError("Cannot divide non-Money by a Money instance.")

    @property
    def is_localized(self):
        if self.use_l10n is None:
            # This definitely raises a warning in Django 4 - we want to ignore RemovedInDjango50Warning
            # However, we cannot ignore this specific warning class as it doesn't exist in older
            # Django versions
            with warnings.catch_warnings():
                warnings.filterwarnings("ignore")
                setting = getattr(settings, "USE_L10N", True)
            return setting
        return self.use_l10n

    def __str__(self):
        format_options = {
            **MONEY_FORMAT,
            **(self.format_options or {}),
        }
        locale = get_current_locale()
        if locale:
            format_options["locale"] = locale
        return moneyed.l10n.format_money(self, **format_options)

    def __html__(self):
        return mark_safe(avoid_wrapping(conditional_escape(str(self))))

    def __round__(self, n=None):
        amount = round(self.amount, n)
        new = self.__class__(amount, self.currency)
        self._copy_attributes(self, new)
        return new

    def round(self, ndigits=0):
        new = super().round(ndigits)
        self._copy_attributes(self, new)
        return new

    def __pos__(self):
        new = super().__pos__()
        self._copy_attributes(self, new)
        return new

    def __neg__(self):
        new = super().__neg__()
        self._copy_attributes(self, new)
        return new

    def __abs__(self):
        new = super().__abs__()
        self._copy_attributes(self, new)
        return new

    def __rmod__(self, other):
        new = super().__rmod__(other)
        self._copy_attributes(self, new)
        return new

    # DefaultMoney sets those synonym functions
    # we overwrite the 'targets' so the wrong synonyms are called
    # Example: we overwrite __add__; __radd__ calls __add__ on DefaultMoney...
    __radd__ = __add__
    __rmul__ = __mul__


def get_current_locale():
    return translation.to_locale(
        translation.get_language()
        # get_language can return None starting from Django 1.8
        or settings.LANGUAGE_CODE
    )


def maybe_convert(value, currency):
    """
    Converts other Money instances to the local currency if `AUTO_CONVERT_MONEY` is set to True.
    """
    if getattr(settings, "AUTO_CONVERT_MONEY", False) and value.currency != currency:
        from .contrib.exchange.models import convert_money

        return convert_money(value, currency)
    return value
